; +-------------------------------------------------------------------------------------------------+
; | The universal source code for commodore floppy disk controllers can compile twelve different    |
; | disk controller versions.                                                                       |
; |                                                                                                 |
; | This is a reverse engineering from the following binaries:                                      |
; |   901466-01 / 901466-02: FDC for DOS 1.x (2040 / 3040 with Shugart drives)                      |
; |   901466-04:             FDC for DOS 2.0 / DOS 2.1 (4040 with Shugart drives)                   |
; |   901483-02:             FDC for DOS 2.5 (8050 with unknown drives, possible Micropolis drives) |
; |   901483-03:             FDC for DOS 2.5 (8050 with Micropolis drives)                          |
; |   901869-01:             FDC for DOS 2.7 (8050 / 8250 with M.P.I. drives)                       |
; |   901884-01:             FDC for DOS 2.7 (8050 / 8250 with Tandon drives)                       |
; |   901885-04:             FDC for DOS 2.7 (8050 / 8250 with Micropolis drives)                   |
; |   251257-02a:            FDC for DOS 2.7 (8250LP / 8296D with Matsushita drives)                |
; |   251474-01b:            FDC for DOS 2.7 (SFD-1001 with Matsushita drives)                      |
; | and some speedos versions (all are based on DOS 2.7):                                           |
; |   FDC for Speedos (8050 / 8250 with Micropolis drives)                                          |
; |   FDC for Speedos (8250LP / 8296D with Matsushita drives, two revisions)                        |
; |                                                                                                 |
; | At this time the DOS 2.5 binaries for Tandon drives are missing. So this source code is not     |
; | included. Also some Speedos versions for Tandon and M.P.I. drives are missing and the version   |
; | for the SFD-1001.                                                                               |
; |                                                                                                 |
; | Also the format (and init) code is assembled for the following binaries:                        |
; |   901468-06:             First half of BC for DOS 1.x (2040 / 3040 with Shugart drives)         |
; |   901468-11 / 901468-14: First third of BC for DOS 2.0 / DOS 2.1 (4040 with Shugart drives)     |
; |   901482-03:             First half of BC for DOS 2.5 (8050 with Micropolis or Tandon drives)   |
; |   901482-06:             First half of BC for DOS 2.5 revisited (8050)                          |
; |   901887-01:             First half of BC for DOS 2.7 (8050 / 8250 / 8250LP / 8296D / SFD-1001) |
; |                                                                                                 |
; | The format and init code from the Speedos versions is at the FDC ROM contained. So the space in |
; | the BC ROM is left.                                                                             |
; | Some labels are used from the Complete Commodore Inner Space Anthology by Karl Hildon, the      |
; | editor of The Transactor magazine (page 53).                                                    |
; |                                                                                                 |
; | (C) 2014 by Martin Hoffmann-Vetter - Germany, All rights reverved! (martinHV@arcor.de)          |
; | The binary code is copyrighted by Commodore Bussines Machines (CBM).                            |
; +-------------------------------------------------------------------------------------------------+

            .debuginfo +

    .if .def(DOS25B) || .def(DOS27)
DOS25       := 1
    .endif

; ===> Constant definition <===

GOODJ       =  1                   ; (00) Job completed successfully
NOHDR       =  2                   ; (20) Header Block not found
NOSYNC      =  3                   ; (21) No Sync character
NODBLK      =  4                   ; (22) Data Block not found
BADBCH      =  5                   ; (23) Data Block checksum eror
WVERER      =  7                   ; (25) Verify error after write
WRPROT      =  8                   ; (26) Write protect on (same as PB_DRV_WPS)
BADHCH      =  9                   ; (27) Header Block checksum error
BADID       = 11                   ; (29) ID mismatch error
BADFMT      = 12                   ; (??) Format error
BADBYT      = 16                   ; (24) Byte decoding error (same as IFR_CB1)

BID         = $07                  ; Data Block ID
HBID        = $08                  ; Header ID Byte
OFFBYT      = $0F                  ; OFF Byte (between header and data block)

READJ       = $00                  ; Read a sector
WRITEJ      = $10                  ; Write a sector
VERFYJ      = $20                  ; Do wverfy
SEEKJ       = $30                  ; Do seek
BUMPJ       = $40                  ; Do bump
JUMPJ       = $50                  ; Do jump (Drive# is ignored)
EXECJ       = $60                  ; Do exec (with motor on)
FORMATJ     = $70                  ; Do format (Only used by Speedos)

; ===> Memory and io definition <===

; Addresses from RIOT

RIOT_RAM    = $0000                ; The 64 bytes RAM from the RRIOT starts at $0000
RIOT        = $0080                ; The IO register from the RRIOT starts at $0080
RIOT_DOUT   = RIOT+$0              ; Data Register A
RIOT_EOUT   = RIOT+$1              ; Data Direction Register A
RIOT_PB     = RIOT+$2              ; Data Register B
RIOT_DDRB   = RIOT+$3              ; Data Direction Register B
RIOT_T8     = RIOT+$F              ; Write Timer /1024

; Control port B

PB_DRV_SEL  = %00000001            ; Drive select (0 or 1)
PB_DRV_DS   = %00000110            ; Drive density select (00, 01, 10 or 11)
PB_DRV_WPS  = %00001000            ; Drive write protect switch
PB_HD_SEL   = %00010000            ; Drive head select (Only used in DOS 2.7 for 8250, SFD-1001 or 8296D)
                                   ; (DOS 2.7 detected a connection to GND for DS drives, else SS drives)
PB_SYNC     = %01000000            ; (Only used in DOS 2.1 and earlier for 4040)
                                   ; (DOS 2.7 detected a connection to GND for 2A format, else 2C format)

; Addresses from VIA

VIA         = $0040                ; The VIA starts at $0040
VIA_VB      = VIA+$0               ; Data Register B
VIA_DIN     = VIA+$1               ; Data Register A
VIA_VDDRB   = VIA+$2               ; Data Direction Register B
VIA_T1LL    = VIA+$4               ; Timer 1 Counter (low)
VIA_TIMER   = VIA+$5               ; Timer 1 Counter (high)
VIA_ACR     = VIA+$B               ; Auxiliary Control Register
VIA_PCR     = VIA+$C               ; Peripheral Control Register
VIA_IFR     = VIA+$D               ; Interrupt Flag Register
VIA_IER     = VIA+$E               ; Interrupt Enable Register

; Control with Peripheral Control Register CA1, CA2, CB1 and CB2

PCR_CA1_NEG = %00000000            ; Detect negative edge on CA1 (Ready)
PCR_CA2_PLS = %00001010            ; Output CA2 puls (Mode Select = ?, ???)
PCR_CA2_GCR = %00001100            ; Output CA2 low (Mode Select = 0, GCR)
PCR_CA2_RAW = %00001110            ; Output CA2 high (Mode Select = 1, RAW)
PCR_CB1_POS = %00010000            ; Detect positive edge on CB1 (Error)
PCR_SEL_WR  = %11000000            ; Output CB2 low (R/W Select = 0, Write)
PCR_SEL_RD  = %11100000            ; Output CB2 high (R/W Select = 1, Read)

PCR_WR_PLS  := PCR_SEL_WR+PCR_CB1_POS+PCR_CA2_PLS+PCR_CA1_NEG ; Write ??? data
PCR_WR_GCR  := PCR_SEL_WR+PCR_CB1_POS+PCR_CA2_GCR+PCR_CA1_NEG ; Write GCR coded data
PCR_WR_RAW  := PCR_SEL_WR+PCR_CB1_POS+PCR_CA2_RAW+PCR_CA1_NEG ; Write raw data
PCR_RD_GCR  := PCR_SEL_RD+PCR_CB1_POS+PCR_CA2_GCR+PCR_CA1_NEG ; Read GCR coded data

; Control port B

VB_TRK_S1A  = %00000001
VB_TRK_S1B  = %00000010
VB_TRK_S0A  = %00000100
VB_TRK_S0B  = %00001000
VB_MOTOR1   = %00010000            ; Drive motor enable for drive 1
VB_MOTOR0   = %00100000            ; Drive motor enable for drive 0
VB_PLLSYN   = %01000000            ;(Only used in DOS 2.5 and later for 8050/8250)
VB_SYNC     = %10000000            ;(Only used in DOS 2.5 and later for 8050/8250)

; Control interrupts

IFR_CA2     = %00000001            ; CA2 Active Edge
IFR_CA1     = %00000010            ; CA1 Active Edge
IFR_SR      = %00000100            ; Complete 8 Shifts
IFR_CB2     = %00001000            ; CB2 Active Edge
IFR_CB1     = %00010000            ; CB1 Active Edge
IFR_T2      = %00100000            ; Time-Out of Timer 2
IFR_T1      = %01000000            ; Time-Out of Timer 1
IFR_IRQ     = %10000000            ; Any enabled Interrupt

; ===> Zeropage declaration <===

            .include "zeropage.s"

; ===> Shared data declaration <===

            .include "shared.s"

; ===> Here starts the format and init code executed from a buffer 0 ($0500) <===

            .setcpu "6502"
            .macpack longbranch
            .segment "DOS"

    .ifndef SPEEDOS                ; In all non-Speedos versions the format code is excluded to the BC ROM
     .ifdef DOS1                   ; Filler to better compare the assembly result with the original for DOS 1.x
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
            .byte 0, 0, 0, 0, 0, 0, 0, 0
     .endif

            .org $0500
            .include "format.s"

     .ifdef DOS25                  ; Beginning with DOS 2.5 the init code for the DC is excluded to the BC ROM
            .org $0500
            .include "init.s"
     .endif

            .reloc
    .endif

; ===> Here starts the FDC code contained in the DC rom <===

            .segment "FDC"

    .ifdef SPEEDOS                 ; The speedos extension use a 4 kbytes rom instead of a 2 kbytes rom from the RRIOT
            .include "speedos.s"
    .endif

    .ifndef DOS1                   ; Beginning with DOS 2.0 first there are two vectors for programm extensions
; ===> Vector to main idling loop for disk controller <===

STARTI:     .word START            ; DC scanning job queue

; ===> Vector to disk controller job completed <===

DONEI:      .word ERROR            ; Terminate disk controller routines
    .endif

    .ifdef DOS25                   ; Beginning with DOS 2.5 the semaphore is directly set
RESET:      ldy #$02
LFC06:      sty TICK
    .endif

    .ifndef DOS1                   ; Beginning with DOS 2.0 the FDC waits until the BC is initialized
LFC04:      lda TICK
            bne LFC04
    .endif

    .ifdef DOS25                   ; Beginning with DOS 2.5 the init code for the DC is excluded to the BC ROM
            dey
            bne LFC06
            jmp INIT
    .else
RESET:      .include "init.s"
    .endif

; ===> The normal main FDC code begin here <===

    .ifdef DOS27
L014:       dey
            bpl L010
    .endif

; ===> DC scanning job queue, waiting for something to do <===

START:      ldy #14                ; A0 0E
    .ifdef DOS25
     .if !(.def(DOS25B) || .def(DOS27))
            lda #$07
            sta BI
     .endif
    .endif
L010:       cli                    ; 58
            lda JOBS,Y             ; B9 03 04
            bpl L014               ; 10 28
    .if .def(SFD) || .def(SPEEDOS)
            sty JOBNUM
    .endif
    .ifndef DOS1
            cmp #%10000000+JUMPJ
     .ifdef DOS25
            bne CONT30
      .if .def(SFD) || .def(SPEEDOS)
            jmp EX
      .else
            tya
       .ifdef DOS27
            sta JOBNUM
       .endif
            jmp LFCCC
      .endif

CONT30:     sei
     .else
            bne LFC65
            sty JOBNUM
            jmp EX
LFC65:
     .endif
    .endif

; ===> Mask out drive number, test if motor running <===

    .ifdef SFD
            jsr LFFEAa             ; 'and #%00000001' and 'tax', if not equal zero (drive 1) then error 03
    .else
            and #%00000001         ; 29 01
            tax                    ; AA
    .endif
    .ifdef SPEEDOS
            nop
            nop
            nop
            nop
            nop
            nop
            nop
    .endif
            sta DRIVE              ; 85 12
    .ifndef DOS25
            sei                    ; 78
    .endif
            lda VIA_VB             ; A5 40
            and ANDC,X             ; 3D ED FF
            beq L012               ; F0 0E

    .ifdef DOS25
            eor VIA_VB
    .else
            lda VIA_VB             ; A5 40
     .ifdef DOS1
            eor ANDE,X             ; 5D F1 FF
     .else
            eor ANDC,X
     .endif
    .endif
            sta VIA_VB             ; 85 40
            lda MTRCLK
    .ifdef SPEEDOS
            clc
    .endif
            adc DELAY              ; 6D 01 04
    .ifdef SPEEDOS
            sta L000A,X
            lda MTRCLK
            clc
            adc #2
    .endif
            sta MTRTM,X            ; 95 01
    .ifdef DOS25
            lda DRVST,X
     .ifdef SPEEDOS
            ora #ST_SPEED_OK+ST_MOTOR_ON
     .else
            ora #ST_SPEED_OK
     .endif
            sta DRVST,X
    .endif

; ===> Wait for motor turned on and head positioned (irq) <===

L012:       lda DRVST,X            ; B5 03
    .ifdef DOS27
            bmi L014
    .else
            bmi L013               ; 30 06
    .endif
    .ifdef SFD
            lda MTRCLK
            sta MTRTM,X
            dec MTRTM,X
    .else
     .ifdef DOS25
            jsr LFFC6
     .else
            lda MTRCLK
            sta MTRTM,X            ; 95 01
     .endif
    .endif
            lda DRVST,X            ; B5 03
L013:       asl A                  ; 0A
    .ifdef DOS27
            bmi L014
    .else
            bpl QUE                ; 10 05

; ===> Decrement job queue pointer <===

L014:       dey                    ; 88
            bpl L010               ; 10 CF
     .ifdef DOS25
            bmi START
     .else
END:        bne START              ; D0 CB
     .endif
    .endif

; ===> Loop to check job queue <===

QUE:        cli                    ; 58
    .ifdef SPEEDOS
            ldy #$FF
            sty NXTJOB
            sty NXTRK
    .endif
    .ifndef DOS25
            lda #$40               ; A9 40
            sta WORK               ; 85 08
    .endif
            ldy #14                ; A0 0E
    .ifndef DOS1
            sty JOBNUM
    .endif

; ===> Check if job for buffer (bit 7 = 1) <===

L020:
    .ifndef DOS1
            jsr SETJB
    .else
            lda JOBS,Y             ; B9 03 04
            sty JOBNUM             ; 84 1A
    .endif
    .ifdef DOS25
     .ifdef SPEEDOS
            bpl LFCB3
            eor DRIVE
            lsr A
            bcs LFCB3
            lda DRVST,X
            and #ST_MOTOR_ON
            beq LFC8D
            lda JOB
            cmp #WRITEJ
            beq LFC8A
            cmp #EXECJ
            beq LFC8A
            cmp #FORMATJ
            bne LFC8D
LFC8A:      lda #%10000000
            .byte $2C
LFC8D:      lda #%00000000
            sta WORK
            jsr LFD29
            bcs LFC9A
            eor #255
            adc #1
LFC9A:      tay
            lda WORK
            eor NXTJOB
            bpl LFCA7
            bit NXTJOB
            bmi LFCAB
            bpl LFCB3
LFCA7:      cpy NXTRK
            bcs LFCB3
LFCAB:      sty NXTRK
            lda WORK
            ora JOBNUM
            sta NXTJOB
LFCB3:      dec JOBNUM
            bpl L020
            lda NXTJOB
            and #%01111111
            sta JOBNUM
            jsr SETJB
            and #%00000010
            lsr a
            sta STAB_ID1
            inc STAB_ID1
            jsr LFD29
            bne LFCD2a
            bit NXTJOB
            bpl GOTU
            bmi LFC7E

LFCD2a:
     .else
            bmi QUE20
QUE05:      dec JOBNUM
            bpl L020
            ldy NXTJOB
            jsr SETJB_Y
            lda NXTRK
     .endif
            sei
            sta STEPS,X
     .ifdef DOS27                  ; For DOS 2.7
      .if .def(PANA) || .def(TANDON) ; Matsushita (Panasonic) or Tandon
            lda #8
            sta STPCNT,X
      .elseif .def(MPI)            ; MPI
            nop
            nop
            nop
            nop
      .else                        ; Micropolis
            lda #4
            sta STEPS2,X           ; 4 Steps for 1 track
      .endif
            lda #ST_HEAD_STP+1
     .else                         ; For DOS 2.5 and earlier
            lda STPTRK
            sta STPCNT,X
            lda #ST_HEAD_STP+0
     .endif
            ora DRVST,X
            sta DRVST,X
            lda (HDRPNT),Y
            sta DRVTRK,X
LFC7E:      jmp START

QUE20:
    .else
            bpl L022               ; 10 27
    .endif
    .ifndef SPEEDOS
            and #%00000001         ; 29 01
            cmp DRIVE              ; C5 12
     .ifdef DOS25
            bne QUE05
      .ifdef DOS27
            lda RIOT_PB
            ora #PB_HD_SEL
            sta RIOT_PB
            lda (HDRPNT),Y
            cmp TAB3
            bcc LFCA2
            lda RIOT_PB
            and #255-PB_HD_SEL
            sta RIOT_PB
            lda DRVTRK,X
            cmp TAB3
            bcs LFCAE
            adc MINTRK_SIDE2
            bne LFCAC
LFCA2:      lda DRVTRK,X
            cmp TAB3
            bcc LFCAE
            sbc MINTRK_SIDE2
LFCAC:      sta DRVTRK,X
LFCAE:
      .endif
            lda DRVTRK,X
     .else
            bne L022               ; D0 21
      .ifndef DOS1
            lda JOB
            cmp #BUMPJ
            bne LFCAF
            jmp BUMP

; ===> Calculate distance to next track <===

LFCAF:
      .endif
            lda DRVST,X            ; B5 03
            and #%00111111         ; 29 3F
            sta TRACK              ; 85 13
     .endif
            beq GOTU               ; F0 41
            sec                    ; 38
     .ifdef DOS1
            sbc TRKS,Y             ; F9 12 04
     .else
            sbc (HDRPNT),Y
     .endif
            beq GOTU               ; F0 3B
    .endif
    .ifdef DOS25
     .ifndef SPEEDOS
            sta NXTRK              ; 85 29
            lda JOBNUM             ; A5 20
            sta NXTJOB             ; 85 28
            bpl QUE05              ; 10 C9
     .endif
GOTU:       ldx #4                 ; A2 04
            lda (HDRPNT),Y         ; B1 19
            sta TRACK              ; 85 14
     .ifdef DOS27
            cmp TAB3
            bcs LFCD2
     .endif
L032:       cmp TAB3-1,X           ; DD AF 04
    .else
            sta DTRCK              ; 85 0A
            bpl LFCA7              ; 10 05
            clc                    ; 18
            eor #255               ; 49 FF
            adc #1                 ; 69 01
LFCA7:      cmp WORK               ; C5 08
            bcs L022               ; B0 06
            sta WORK               ; 85 08
            lda DTRCK              ; A5 0A
            sta WORK+1             ; 85 09
     .ifdef DOS1
L022:       dey                    ; 88
     .else

; ===> Check other jobs, bump head <===

L022:       dec JOBNUM
     .endif
            bpl L020               ; 10 CF
            lda WORK               ; A5 08
            bit WORK+1             ; 24 09
            bmi LFCBF              ; 30 05
            clc                    ; 18
            eor #255               ; 49 FF
            adc #1                 ; 69 01
LFCBF:      sta WORK               ; 85 08
            asl A                  ; 0A
            sei                    ; 78
            sta STEPS,X            ; 95 05
            lda #ST_HEAD_STP       ; A9 40
            ora DRVST,X            ; 15 03
            clc                    ; 18
            adc WORK               ; 65 08
            sta DRVST,X            ; 95 03
FIN:        bne END                ; D0 AA

     .ifdef DOS1
TAB1:       .byte 17, 18, 20, 21
     .endif

ANDA:       .byte 255-(VB_TRK_S0B+VB_TRK_S0A)

TAB3:       .byte 255-(VB_TRK_S1B+VB_TRK_S1A)
            .byte     31, 25, 18   ; 35 - 32, 31 - 26, 25 - 19, 18 - 1

; ===> Check position of track <===

GOTU:       lda DRVST,X            ; B5 03
            bmi FIN                ; 30 F1
            ldx #4                 ; A2 04
     .ifdef DOS1
            lda TRKS,Y             ; B9 12 04
     .else
            lda (HDRPNT),Y
     .endif

; ===> Determine zone <===

L032:       cmp ANDA,X             ; DD D4 FC
    .endif
            dex                    ; CA
            bcs L032               ; B0 FA
    .ifdef DOS27
            bcc L031

LFCD2:      cmp TAB3+3,X
            dex
            bcs LFCD2
    .endif

L031:       lda TAB1,X             ; BD D0 FC
            sta SECTR              ; 85 15
            txa                    ; 8A
            asl A                  ; 0A
            sta WORK               ; 85 08
            lda RIOT_PB            ; A5 82
            and #255-(PB_DRV_DS+PB_DRV_SEL)
            ora WORK               ; 05 08
            ora DRIVE              ; 05 12
            sta RIOT_PB            ; 85 82
            ldx DRIVE              ; A6 12
    .ifdef DOS1
            lda JOBS,Y             ; B9 03 04
            and #%01110000         ; 29 70
    .else
     .ifdef DOS25
      .ifndef SPEEDOS
            lda DRVST,X
            bmi LFC7E
      .endif
     .endif
            lda JOB
    .endif
    .ifdef DOS25
            cmp #BUMPJ
            beq BUMP
    .endif
EXE:        cmp #EXECJ             ; C9 60
            beq EX                 ; F0 07
    .ifdef DOS1
            cmp #BUMPJ             ; C9 40
            beq BUMP               ; F0 0C
    .endif
    .ifdef SPEEDOS
            cmp #FORMATJ
            bne LFD26
            jmp FORMT
LFD26:
    .endif
            jmp SEEK               ; 4C 65 FE

    .ifdef SPEEDOS
LFD29:      lda RIOT_PB
            ora #PB_HD_SEL
            sta RIOT_PB
            lda (HDRPNT),y
            cmp TAB3
            bcc LFD48
            lda RIOT_PB
            and #255-PB_HD_SEL
            sta RIOT_PB
            lda DRVTRK,x
            cmp TAB3
            bcs LFD54
            adc MINTRK_SIDE2
            bne LFD52
LFD48:      lda DRVTRK,x
            cmp TAB3
            bcc LFD54
            sbc MINTRK_SIDE2
LFD52:      sta DRVTRK,x
LFD54:      lda DRVTRK,x
            sec
            beq LFD5B
            sbc (HDRPNT),y
LFD5B:      rts
    .endif

; ===> Execute job in data buffer <===

    .ifdef DOS1
EX:         tya                    ; 98
    .else
EX:         lda JOBNUM
    .endif
LFCCC:      clc                    ; 18
            adc #>BUFF0            ; 69 05
            sta BUFPT+1            ; 85 17
            jmp (BUFPT)            ; 6C 16 00

; ===> Position head (bump to track) <===

BUMP:
    .ifdef SPEEDOS
            jsr LFD6C
            jmp FMTEND

LFD6C:      sei
            lda ANDD,X
            ora VIA_VB
            sta VIA_VB

            lda #1
            sta DRVTRK,X           ; Bump to track 1

            ldy #92                ; 92 for DOS 2.7
LFD7A:
    .endif
            sei                    ; 78
    .if .def(DOS27)
            lda #ST_HEAD_STP+1     ; track 1 for DOS 2.7
    .elseif .def(DOS25)
            lda #ST_HEAD_STP+0     ; track 0 (?) for DOS 2.5
    .else
            lda #ST_SPEED_OK+ST_HEAD_STP+1 ; track 1 for DOS 1.x and 2.0
    .endif
    .ifdef DOS1
            ldx DRIVE              ; A6 12
    .endif
    .ifdef SPEEDOS
            ora DRVST,X
    .endif
            sta DRVST,X            ; 95 03
    .ifdef SPEEDOS
            tya
            sta STEPS,x
            lda L04EE
            sta STPCNT,x
LFD8A:      sei
            jsr LFFC6
            cli
            lda DRVST,x
            asl A
            bmi LFD8A
            sei
            rts
    .else
     .ifdef DOS25
            lda #1
            sta DRVTRK,X           ; BUMP to track 1

            lda ANDD,X
     .else
            lda #VB_TRK_S0A+VB_TRK_S0B+VB_TRK_S1A+VB_TRK_S1B
            and ANDD,X             ; 3D EF FF
     .endif
            ora VIA_VB             ; 05 40
            sta VIA_VB             ; 85 40
     .if .def(DOS27)
            lda #92                ; 92 for DOS 2.7
     .elseif .def(DOS25)
            lda #90                ; 90 for DOS 2.5
     .else
            lda #140               ; 140 for DOS 1.x and 2.0
     .endif
            sta STEPS,X            ; 95 05
     .if .def(PANA) || .def(TANDON)
            jmp LFFEA              ; Patch 'sta STPCNT,X' before 'jmp DONE'
     .else
            jmp DONE               ; 4C C1 FE
     .endif
    .endif

; ===> <===

    .ifdef DOS25
SETJB:      ldy JOBNUM
SETJB_Y:    lda JOBS,Y
            pha
            and #%01111000
            sta JOB
            tya
            asl A
            asl A
            asl A
            adc #<HDRS
            sta HDRPNT
            ldy #$02
            pla
            rts

FSNUM:      ldy #3
            lda (HDRPNT),Y
            sta STAB_SEC
    .endif
    .ifdef SPEEDOS
            lda #$02
    .endif

; ===> <===

WSECT:
    .ifdef SPEEDOS
            ldy SECTR
            sty CSECT
            ldy #3
            sty DSECT
LFDBC:
            cli
            clc
            adc STAB_SEC
            cmp SECTR
            bcc LFDC6
            sbc SECTR
LFDC6:      sta NEXTS
            ldx DRIVE
            lda DRVST,X
            and #ST_MOTOR_ON
            sta NXTRK
    .else
     .ifdef DOS1
            lda #$FF               ; A9 FF
            sta JOBNUM             ; 85 1A
            lsr A                  ; 4A
     .else
            lda #$7F
     .endif
            sta CSECT              ; 85 0C
            lda STAB_SEC           ; A5 10
            clc                    ; 18
     .ifdef DOS1
            adc #$01               ; 69 01
     .else
            adc #$02
     .endif
            cmp SECTR              ; C5 15
     .ifdef DOS1
            bne LFD3F              ; D0 02
     .else
            bcc LFD3F
     .endif
            lda #$00               ; A9 00
LFD3F:      sta NEXTS              ; 85 14
    .endif
            ldx #14                ; A2 0E
    .ifdef DOS1
L480:       lda JOBS,X             ; BD 03 04
            bpl LFD83              ; 10 3B
            lda TRKS,X             ; BD 12 04
            and #%00111111         ; 29 3F
            cmp TRACK              ; C5 13
            bne LFD83              ; D0 32
            lda JOBS,X             ; BD 03 04
            tay                    ; A8
    .else
     .ifdef SPEEDOS
            stx WORK
     .else
            stx JOBNUM
     .endif
            ldx #$FF
L480:
     .ifdef SPEEDOS
            ldy WORK
            jsr SETJB_Y
     .else
            jsr SETJB
     .endif
            bpl LFD83
     .ifdef SPEEDOS
            eor DRIVE
            lsr A
            bcs LFD83
     .else
            sta WORK
     .endif
     .ifndef DOS25
            lda (HDRPNT),Y
            cmp TRACK
            bne LFD83
            LDA WORK            
     .endif
    .endif
    .ifndef SPEEDOS
            and #%00000001         ; 29 01
            cmp DRIVE              ; C5 12
            bne LFD83              ; D0 28
    .endif
    .ifdef DOS25
            lda (HDRPNT),Y
            cmp TRACK
            bne LFD83
            lda JOB
    .else
            tya                    ; 98
    .endif
    .ifdef SPEEDOS
            bit LFFF9
            bne LFDFB
            cmp #SEEKJ
            bcs LFD83
            cmp #WRITEJ
            bne LFDFB
            lda NXTRK
            bne LFD83
LFDFB:      iny
            ldx DSECT
            beq LFE13
            lda (HDRPNT),y
            cmp NEXTS
            beq LFE13
    .else
            cmp #EXECJ             ; C9 60
            beq LFD83              ; F0 23
     .ifdef DOS1
            txa                    ; 8A
            jsr SETJB              ; 20 99 FD
            ldy #3                 ; A0 03
     .else
            iny
     .endif
            sec                    ; 38
            lda (HDRPNT),Y         ; B1 18
            sbc NEXTS              ; E5 14
            bpl LFD70              ; 10 03
            clc                    ; 18
            adc SECTR              ; 65 15
     .ifdef DOS25
LFD70:      cmp CSECT
            bcs LFD83
DOITT:      sta CSECT
            lda JOBNUM
            tax
     .else
LFD70:      sta DSECT              ; 85 0B
            sec                    ; 38
            sbc CSECT              ; E5 0C
            bpl LFD83              ; 10 0C
      .ifdef DOS1
            stx JOBNUM             ; 86 1A
      .else
            ldx JOBNUM
      .endif
            lda DSECT              ; A5 0B
            sta CSECT              ; 85 0C
            txa                    ; 8A
            clc                    ; 18
     .endif
            adc #>BUFF0            ; 69 05
            sta BUFPT+1            ; 85 17
    .endif
LFD83:
    .ifdef DOS1
            dex                    ; CA
    .else
     .ifdef SPEEDOS
            dec WORK
     .else
            dec JOBNUM
     .endif
    .endif
            bpl L480               ; 10 BD
    .ifdef DOS1
            ldy JOBNUM             ; A4 1A
    .else
            txa
    .endif
            bpl HPNT               ; 10 03
            jmp START              ; 4C 47 FC

HPNT:
    .ifdef SPEEDOS
            jmp LFEDD

LFE13:      lda WORK
            sta JOBNUM
            sta ACTJOB
            clc
            adc #>BUFF0
            sta BUFPT+1
            lda JOB
    .else
     .ifdef DOS1
            tya                    ; 98
            jsr SETJB              ; 20 99 FD
            lda JOBS,Y             ; B9 03 04
            and #%01110000         ; 29 70
     .else
            stx ACTJOB
            stx JOBNUM
            jsr SETJB
      .ifdef DOS25
       .if !(.def(DOS25B) || .def(DOS27))
            stx BI
       .endif
      .endif
            lda JOB
     .endif
     .ifndef DOS25
            jmp READ               ; 4C AA FD

; ===> Set pointer in header buffer according to job code <===

SETJB:
      .ifndef DOS1
            ldy JOBNUM
            lda JOBS,Y
            pha
            and #%01110000
            sta JOB
            tya
      .endif
            asl A                  ; 0A
            asl A                  ; 0A
            asl A                  ; 0A
            adc #<HDRS             ; 69 21
            sta HDRPNT             ; 85 18
      .ifndef DOS1
            ldy #$02
            pla
      .endif
            rts                    ; 60

; ===> Optimize job <===

FSNUM:
            ldy #3                 ; A0 03
            lda (HDRPNT),Y         ; B1 18
            sta STAB_SEC           ; 85 10
            jmp WSECT              ; 4C 2D FD
     .endif

; ===> Read sector into buffer (code 00) <===
READ:       cmp #READJ             ; C9 00
    .endif
            bne WRITE              ; D0 28
            jsr DSTRT              ; 20 C3 FD

; ===> Wait for data byte to be ready <===

L100:
    .ifdef DOS25
            bvc L100
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L100               ; 10 FC
    .endif
            lda VIA_DIN            ; A5 41
            sta (BUFPT),Y          ; 91 16
    .ifndef DOS25
            eor WORK               ; 45 08
            sta WORK               ; 85 08
    .endif
            iny                    ; C8
            bne L100               ; D0 F1
    .ifdef DOS25
RDCHKSUM:   bvc RDCHKSUM
            lda VIA_DIN
            sta CHKSUM2
            lda #IFR_CB1
            bit VIA_IFR
            bne LFD8E              ; => Error #BADBYT
            jsr CHKBLK
            cmp CHKSUM2
            bne LFD8C
            jmp DONE
    .else
            jmp DEND               ; 4C 4E FE
    .endif

    .ifdef DOS25
LFD8C:      lda #BADBCH
     .ifdef DOS27
            .byte $2C
LFDC2:      lda #NODBLK
ER:
     .endif
LFD8E:      jmp ERROR
    .endif

; ===> Find a data block ID (07) <===

DSTRT:
    .ifndef DOS25
            ldy #$00               ; A0 00
            sty WORK               ; 84 08
    .endif
            jsr SRCH               ; 20 CE FE
            jsr SYNC               ; 20 3F FF
            cmp #BID               ; C9 07
    .ifdef DOS27
            bne LFDC2
    .else
            beq LFDD5              ; F0 04
            lda #NODBLK            ; A9 04
     .ifdef DOS25
            bne ER
     .else
            bne LFDE2              ; D0 0D
     .endif
    .endif

; ===> Found a goog block <===

LFDD5:      rts                    ; 60

; ===> Verify write protect <===

WRITE:
    .ifdef DOS25
            cmp #WRITEJ
            bne VRFY
            jsr CHKBLK
            sta CHKSUM2
    .else
            cmp #VERFYJ            ; C9 20
            bpl VRFY               ; 10 60
    .endif
            lda RIOT_PB            ; A5 82
            and #PB_DRV_WPS        ; 29 08
    .ifdef DOS27
            bne ER                 ; => Error #PB_DRV_WPS
    .else
            beq L198               ; F0 04
     .ifdef DOS1
            lda #WRPROT            ; A9 08
     .endif
LFDE2:
     .ifdef DOS1
            bne ER                 ; D0 7F
     .else
      .ifdef DOS25
ER:
      .endif
            jmp ERROR
     .endif
    .endif

; ===> Write buffer to disk <===

L198:
    .ifndef DOS25
            lda #IFR_CB1           ; A9 10
            sta VIA_IER            ; 85 4E
    .endif
            jsr SRCH               ; 20 CE FE
    .ifndef DOS1
            ldx GAP1
     .ifndef DOS27
            dex
     .endif
    .endif

; ===> Check OFF bytes <===

L200:
    .ifdef DOS25
            bvc L200
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L200               ; 10 FC
            bit VIA_DIN            ; 24 41
    .endif
    .ifndef DOS1
            dex
            bne L200
    .endif
            lda #PCR_WR_RAW
            sta VIA_PCR            ; 85 4C
    .ifdef DOS1
            ldx #%11111111         ; A2 FF
            stx RIOT_DOUT          ; 86 80
            stx RIOT_EOUT          ; All bits on this port are output
L201:       bit VIA_IFR            ; 24 4D
            bpl L201               ; 10 FC
            bit VIA_DIN            ; 24 41
            stx RIOT_DOUT          ; 86 80
            lda #PCR_WR_GCR
    .else
            lda #PCR_WR_GCR
            ldx #$FF
     .ifdef DOS25
            jsr WRT2BYT
            jsr WRT2BYT
     .else
            jsr OUT                ; Write character in .X to disk
            jsr OUT                ; Write character in .X to disk
            jsr OUT                ; Write character in .X to disk
     .endif
    .endif
L202:
    .ifdef DOS25
            bvc L202
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L202               ; 10 FC
            bit VIA_DIN            ; 24 41
    .endif
            sta VIA_PCR            ; 85 4C
            lda #BID               ; A9 07
            sta RIOT_DOUT          ; 85 80
            ldy #$00               ; A0 00
    .ifdef DOS25
L203:       bvc L203
            clv
    .else
            sty WORK               ; 84 08
L203:       bit VIA_IFR            ; 24 4D
            bpl L203               ; 10 FC
            bit VIA_DIN            ; 24 41
    .endif
            lda (BUFPT),Y          ; B1 16
            sta RIOT_DOUT          ; 85 80
    .ifndef DOS25
            eor WORK               ; 45 08
            sta WORK               ; 85 08
    .endif
            iny                    ; C8
            bne L203               ; D0 EF
    .ifdef DOS25
            ldx CHKSUM2
LFDE4:      bvc LFDE4
            clv
            stx RIOT_DOUT
            jsr WRTBYT
            ldx #$00
            jsr WRTBYT
            lda #PCR_RD_GCR
LFDF3:      bvc LFDF3
            clv
            sta VIA_PCR
            ldx #40
LFDFA:      jsr RDBYT
            dex
            bne LFDFA
    .else
            tax                    ; AA
            jsr OUT                ; Write character in .X to disk
            jsr OFF                ; 20 59 FF
    .endif
            ldy JOBNUM             ; A4 1A
            lda JOBS,Y             ; B9 03 04
            eor #%00110000         ; 49 30
            sta JOBS,Y             ; 99 03 04
            jmp FSNUM              ; 4C A1 FD

; ===> Compare sector with buffer <===

VRFY:
    .ifdef DOS25
            cmp #VERFYJ
            bne SECTSK
    .endif
            jsr DSTRT              ; 20 C3 FD
L210:
    .ifdef DOS25
            bvc L210
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L210               ; 10 FC
    .endif
            lda VIA_DIN            ; A5 41
            cmp (BUFPT),Y          ; D1 16
            bne L212               ; D0 1A
    .ifdef DOS25
            iny
            bne L210
            jmp RDCHKSUM
    .else
            eor WORK               ; 45 08
            sta WORK               ; 85 08
            iny                    ; C8
            bne L210               ; D0 EF
DEND:       jsr BYTE               ; 20 52 FF
            cmp WORK               ; C5 08
            beq L214               ; F0 04
            lda #BADBCH            ; A9 05
            bne ER                 ; D0 0A
L214:       lda #IFR_CB1           ; A9 10
            bit VIA_IFR            ; 24 4D
            beq DONE               ; F0 62
            bne ER                 ; => Error #BADBYT
    .endif

    .if !(.def(SFD) || .def(SPEEDOS))
L212:       lda #WVERER            ; A9 07
     .ifndef DOS25
ER:
     .endif
            bne ERR                ; D0 5E
    .endif
    .ifdef DOS25
SECTSK:     jsr SRCH
     .ifdef DOS27
            beq DONE
     .else
            jmp DONE
     .endif
    .endif

; ===> Read sector header and compare ID <===

SEEK:
    .ifdef SPEEDOS
            sec
            ror DSECT
LFEDD:
    .endif
    .ifdef DOS25
     .if .def(DOS25B) || .def(DOS27)
            ldx #180               ; 180 for Last Revision DOS 2.5 or DOS 2.7 (read 90 sector header)
     .else
            ldx #90                ; 90 for first Revision DOS 2.5 (read 90 sector header)
     .endif
    .else
            lda #$00               ; A9 00
            sta WORK               ; 85 08
            ldx #6                 ; 6 for DOS 1.x and DOS 2.0 (read 6 sector header)
    .endif
            jsr HEAD               ; 20 F3 FE
    .ifdef SPEEDOS
            ldx #4
    .else
            ldy #4                 ; A0 04
    .endif
L250:
    .ifdef DOS25
            bvc L250
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L250               ; 10 FC
    .endif
            lda VIA_DIN            ; A5 41
    .ifdef SPEEDOS
            sta STAB,X
    .else
            sta STAB,Y             ; 99 0D 00
    .endif
    .ifdef DOS25
     .ifdef SPEEDOS
            dex
     .else
            dey
     .endif
            bpl L250
     .ifdef SPEEDOS
            lda #IFR_CB1
            and VIA_IFR
            bne LFEFE              ; => Error #BADBYT if no retry
            ldx #$04
LFE43:      eor STAB,X
            dex
     .else
            ldy #4
            lda #$00
LFE43:      eor STAB,Y
            dey
     .endif
            bpl LFE43
    .else
            eor WORK               ; 45 08
            sta WORK               ; 85 08
            dey                    ; 88
            bpl L250               ; 10 F0
    .endif
    .ifdef SPEEDOS
LFEFE:      bit DSECT
            bpl LFF22
            tax
    .else
            cmp #$00               ; C9 00
    .endif
            bne CSERR              ; D0 42
    .ifndef DOS25
            lda STAB_TRK           ; A5 0F
    .endif
    .ifdef DOS1
            and #%00111111         ; 29 3F
    .endif
            ldx DRIVE              ; A6 12
    .ifdef DOS25
            lda STAB_TRK
            sta DRVTRK,X
    .else
            sta DRVST,X            ; 95 03
    .endif
    .ifdef DOS1
            ldy JOBNUM             ; A4 1A
            lda JOBS,Y             ; B9 03 04
            and #%01110000         ; 29 70
    .else
            lda JOB
    .endif
            cmp #SEEKJ             ; C9 30
            beq ESEEK              ; F0 1C
    .ifdef DOS1
            tya                    ; 98
            jsr SETJB              ; 20 99 FD
            ldy #2                 ; A0 02
            lda (HDRPNT),Y         ; B1 18
            eor STAB_TRK           ; 45 0F
            and #%11000000         ; 29 C0
            bne L253               ; D0 25
            dey                    ; 88
    .else
            ldy #1
    .endif

; ===> Compare block header ID with current (disk) ID <===

L252:       lda (HDRPNT),Y         ; B1 18
            cmp STAB,Y             ; D9 0D 00
            bne L253               ; D0 1D
            dey                    ; 88
            bpl L252               ; 10 F6
    .ifdef SPEEDOS
            lda #$01
    .endif
            jmp WSECT              ; 4C 2D FD

    .ifdef SPEEDOS
LFF22:      tax
            beq LFF2D
            lda NEXTS
            sta STAB_SEC
            dec DSECT
            bmi LFF35
LFF2D:      lda DSECT
            beq LFF39
            dec CSECT
            bpl LFF39
LFF35:      lda #$00
            sta DSECT
LFF39:      lda #$01
            jmp LFDBC
    .endif

; ===> Transfer sector header to header buffer <===

ESEEK:
    .ifdef DOS1
            tya                    ; 98
            jsr SETJB              ; 20 99 FD
    .endif
            ldy #4                 ; A0 04
L251:       lda STAB,Y             ; B9 0D 00
            sta (HDRPNT),Y         ; 91 18
            dey                    ; 88
            bpl L251               ; 10 F8

; ===> DC message 1, OK (BC: 00) <===

DONE:       lda #GOODJ             ; A9 01
    .ifdef DOS25
            .byte $2C

; ===> DC message 9, header checksum error (BC: 27) <===

CSERR:      lda #BADHCH
            .byte $2C

     .if .def(SFD) || .def(SPEEDOS)
L212:       lda #WVERER
            .byte $2C
     .endif

; ===> DC message 11, wrong disk ID (BC: 29) <===

L253:       lda #BADID
    .endif
ERR:        jmp ERROR              ; 4C F9 FE
    .ifndef DOS25

; ===> DC message 9, header checksum error (BC: 27) <===

CSERR:      lda #BADHCH            ; A9 09
            bne ERR                ; D0 F9

; ===> DC message 11, wrong disk ID (BC: 29) <===

L253:       lda #BADID             ; A9 0B
            bne ERR                ; D0 F5
    .endif

; ===> Calculate block header checksum and put in header buffer <===

SRCH:
    .ifdef SPEEDOS
            ldy #0
            tya
    .else
            ldy #3                 ; A0 03
            lda #$00               ; A9 00
    .endif
L412:       eor (HDRPNT),Y         ; 51 18
    .ifdef SPEEDOS
            iny
            cpy #4
            bcc L412
    .else
            dey                    ; 88
            bpl L412               ; 10 FB
            ldy #4                 ; A0 04
    .endif
            sta (HDRPNT),Y         ; 91 18
    .ifndef SPEEDOS
            ldy JOBNUM             ; A4 1A
    .endif
    .if .def(DOS25B) || .def(DOS27)
            ldx #180               ; 180 for last Revision DOS 2.5 or DOS 2.7
    .else
            ldx #90                ; 90 else
    .endif

; ===> Compare next header with header buffer <===

L410:       jsr HEAD               ; 20 F3 FE
            ldy #4                 ; A0 04
L411:
    .ifdef DOS25
            bvc L411
            clv
    .else
            bit VIA_IFR            ; 24 4D
            bpl L411               ; 10 FC
    .endif
            lda VIA_DIN            ; A5 41
            cmp (HDRPNT),Y         ; D1 18
            bne L410               ; D0 F1
            dey                    ; 88
            bpl L411               ; 10 F3
            iny                    ; C8
    .if !(.def(DOS1) || .def(DOS25))
            lda VIA_VB
            ora #VB_SYNC
            sta VIA_VB
            and #255-VB_SYNC
            sta VIA_VB
    .endif
LFECDa:     rts                    ; 60

    .ifdef DOS27
; ===> Find start of sector header (08) <===

L420:       jsr SYNC
            cmp #HBID
            beq LFECDa
    .endif

; ===> Find next sector header <===

HEAD:       cli                    ; 58
            dex                    ; CA
    .ifdef DOS25
            bne L420
    .else
            bpl L420               ; 10 2E
    .endif

            lda #NOHDR             ; A9 02

; ===> Terminate disk controller routines <===

ERROR:
    .ifdef DOS1
            bit JOB                ; 24 1E
            bpl ERR1               ; 10 0A
            inc ERRCNT             ; E6 1F
            ldx #$3F               ; A2 3F
            txs                    ; 9A
            ldx DRIVE              ; A6 12
            jmp L0543              ; 4C 43 05

ERR1:
    .endif
            ldy JOBNUM             ; A4 1A
    .ifdef SPEEDOS
            pha
            and #255-%10000000
    .endif
            sta JOBS,Y             ; 99 03 04
    .ifndef DOS1
     .ifndef SPEEDOS
            pha
     .endif
    .endif
    .ifdef DOS25
            lda MTRCLK
    .endif
            ldx DRIVE              ; A6 12
    .ifndef DOS25
            lda MTRCLK
    .endif
            adc CUTMT              ; 6D 02 04
            sta MTRTM,X            ; 95 01
    .ifdef DOS25
     .ifndef DOS27
            jsr LFF0B
     .endif
    .endif
    .ifdef DOS1
            lda JOBS,Y             ; B9 03 04
            cmp #GOODJ             ; C9 01
    .else
            pla
            lsr A
    .endif
            bne L421               ; D0 03
            jmp FSNUM              ; 4C A1 FD

; ===> Abort after erro <===

L421:       ldx #$3F               ; A2 3F
            txs                    ; 9A
            jmp START              ; 4C 47 FC

    .ifndef DOS27
; ===> Find start of sector header (08) <===

L420:       jsr SYNC               ; 20 3F FF
            cmp #HBID              ; C9 08
            bne HEAD               ; D0 C7
            rts                    ; 60
    .endif

    .ifndef DOS25
WATCH:      bit RIOT_PB            ; 24 82
            bvc LFF3E              ; 50 0D
            bit VIA_IFR            ; 24 4D
            bpl WATCH              ; 10 F8
            bit VIA_DIN            ; 24 41
            bit VIA_VB             ; 24 40
            dex                    ; CA
            bne WATCH              ; D0 F1
            bit RIOT_PB            ; 24 82
LFF3E:      rts                    ; 60
    .endif

; ===> Wait for a data block sync mark <===

SYNC:
    .ifdef DOS25
     .if .def(DOS25B) || .def(DOS27)
            lda #$1E               ; 30 ???
            sta SYNCCNT
     .endif
LFECD:
     .ifndef SPEEDOS
      .ifdef DOS27
            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB
      .else
            jsr LFF0B
      .endif
     .endif
    .endif
            sei                    ; 78
    .ifdef SPEEDOS
            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB
    .endif
            lda #>$D000            ; 53248 us
            sta VIA_TIMER          ; 85 45
            lda #NOSYNC            ; A9 03
    .if .def(DOS25B) || .def(DOS27)
            dec SYNCCNT
            bmi ERROR
    .endif

; ===> Read timer, test sync <===

LFF46:      bit VIA_TIMER          ; 24 45
            bpl ERROR              ; 10 AF
    .ifdef DOS25
            bit VIA_VB
            bmi LFF46
            ldy SYNDLY
LFEE3:      bit VIA_VB
     .if .def(DOS25B) || .def(DOS27)
            bmi LFECD
     .else
            bmi SYNC
     .endif
            dey
            bne LFEE3
            lda VIA_VB
     .ifdef SPEEDOS
LFFCB:
     .endif
            ora #VB_PLLSYN
            sta VIA_VB
            bit VIA_DIN
            clv
RDBYT:      bvc RDBYT
            clv
    .else
            bit RIOT_PB            ; 24 82
            bvs LFF46              ; 70 F8
            bit VIA_VB             ; 24 40
            bit VIA_DIN            ; 24 41
BYTE:       bit VIA_IFR            ; 24 4D
            bpl BYTE               ; 10 FC
    .endif
            lda VIA_DIN            ; A5 41
            rts                    ; 60

    .ifdef DOS25
WRT2BYT:    jsr WRTBYT
WRTBYT:     bvc WRTBYT
            clv
            stx RIOT_DOUT
            rts

CHKBLK:     lda #$00
            tay
LFF05:      eor (BUFPT),Y
            iny
            bne LFF05
            rts

     .ifndef DOS27
LFF0B:      lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB
            rts
     .endif
    .else

; ===> Write byte to disk <===

OFF:        jsr OUT                ; Write character in .X to disk
            ldx #$00               ; A2 00
            jsr OUT                ; Write character in .X to disk
            lda #PCR_RD_GCR

; ===> Read <===

L442:       bit VIA_IFR            ; 24 4D
            bpl L442               ; 10 FC
     .ifdef DOS1
            stx RIOT_EOUT          ; 86 81
     .endif
            sta VIA_PCR            ; 85 4C
     .ifdef DOS1
            lda #IFR_IRQ+IFR_CB1+IFR_CA1+IFR_CB2
     .else
            lda #IFR_IRQ+IFR_CB1+IFR_CA1
     .endif
            sta VIA_IER            ; 85 4E
            jsr BYTE               ; 20 52 FF
            bit VIA_VB             ; 24 40
            jsr BYTE               ; 20 52 FF
            bit VIA_VB             ; 24 40
            jmp BYTE               ; 4C 52 FF

; ===> Write character in .X to disk <===

OUT:        bit VIA_IFR            ; 24 4D
            bpl OUT                ; 10 FC
            stx RIOT_DOUT          ; 86 80
            bit VIA_DIN            ; 24 41
            rts                    ; 60
    .endif

.ifndef SPEEDOS
            .include "irqh.s"
.endif

    .ifdef DOS25
     .ifndef SFD
LFFC6:      lda MTRCLK
            sta MTRTM,X
            dec MTRTM,X
            rts
     .endif

ANDA:       .byte 255-(VB_TRK_S0A+VB_TRK_S0B)
            .byte 255-(VB_TRK_S1A+VB_TRK_S1B)
    .endif

    .if .def(MPI) || .def(PANA) || .def(TANDON)
ANDB:       .byte VB_TRK_S0B            ; Step bits drive 0
            .byte VB_TRK_S1B            ; Step bits drive 1

     .ifdef SPEEDOS
      .ifdef REVB
ANDC        := $FCB4
LFFF5:      .byte 255-VB_MOTOR0         ; Acceleraton/Deceleration bits drive 0
            .byte 255-VB_MOTOR1         ; Acceleraton/Deceleration bits drive 1
      .else
ANDC:       .byte VB_MOTOR0             ; Acceleraton/Deceleration bits drive 0
            .byte VB_MOTOR1             ; Acceleraton/Deceleration bits drive 1
      .endif
     .endif

ANDD:       .byte VB_TRK_S0A+VB_TRK_S0B ; Motor timing bits 0
            .byte VB_TRK_S1A+VB_TRK_S1B ; Motor timing bits 1
    .else

ANDB:       .byte VB_TRK_S0A            ; Step bits drive 0
            .byte VB_TRK_S1A            ; Step bits drive 1

     .ifdef SPEEDOS
ANDC:       .byte VB_MOTOR0             ; Acceleraton/Deceleration bits drive 0
            .byte VB_MOTOR1             ; Acceleraton/Deceleration bits drive 1
     .endif
    .endif

    .ifdef SPEEDOS
     .if !(.def(MPI) || .def(PANA) || .def(TANDON))
ANDD:       .byte VB_TRK_S0A+VB_TRK_S0B ; Motor timing bits 0
            .byte VB_TRK_S1A+VB_TRK_S1B ; Motor timing bits 1
     .endif

LFFF9:      .byte %00001000
    .else

ANDC:       .byte VB_MOTOR0             ; Acceleraton/Deceleration bits drive 0
            .byte VB_MOTOR1             ; Acceleraton/Deceleration bits drive 1

     .if !(.def(MPI) || .def(PANA) || .def(TANDON))
ANDD:       .byte VB_TRK_S0A+VB_TRK_S0B ; Motor timing bits 0
            .byte VB_TRK_S1A+VB_TRK_S1B ; Motor timing bits 1
     .endif
    .endif

    .ifdef DOS1
            ; What does this doing? %10000000 and %01000000 are PB7 (Pin 17) and PB6 (Pin 16) from VIA and not connected!
ANDE:       .byte %10000000+VB_MOTOR0
            .byte %01000000+VB_MOTOR1
    .endif

    .if .def(DOS27)                ; DOS 2.7
     .if .def(SPEEDOS)             ; Speedos
;            .res 1
FILLER:     .byte $AA

ROMCS:
      .if .def(PANA)               ; Matsushita (Panasonic)
       .ifdef REVB
            .byte $5A              ; Checksum
       .else
            .byte $59              ; Checksum
       .endif
      .elseif .def(MPI)            ; MPI
;ToDo: Unknown!!!
      .else                        ; Micropolis
            .byte $BF              ; Checksum
      .endif
     .elseif .def(PANA) || .def(TANDON) ; Matsushita (Panasonic) or Tandon
      .ifndef SFD
;            .res 2
            .byte $AA, $AA
      .endif
LFFEA:      sta STPCNT,X
            jmp DONE

      .ifdef SFD
LFFEAa:     and #%00000001
            tax
            bne LFFF0
            rts
LFFF0:      lda #NOSYNC
            jmp ERROR

            ; DOS 2.7 - 251474-01b (SFD-1001 with Matsushita drives)
;            .res 6
FILLER:     .byte $AA, $AA, $AA, $AA, $AA

LFFFA:      .byte $AA              ; Checksum 2
ROMCS:      .byte $09              ; Checksum 1
      .else
            ; DOS 2.7 - 251257-02a (8250LP or 8296D with Matsushita drives)
            ; DOS 2.7 - 901884-01 (8050 or 8250 with Tandon drives)
;            .res 11
FILLER:     .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA

LFFFA:      .byte $53              ; Checksum 2
ROMCS:      .byte $A3              ; Checksum 1 (same as DOS 2.7 for MPI drives)
      .endif
     .elseif .def(MPI)             ; MPI
            ; DOS 2.7 - 901869-01 (8050 or 8250 with MPI drives)
;            .res 19
FILLER:     .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA

LFFFA:      .byte $AA              ; Checksum 2
ROMCS:      .byte $A3              ; Checksum 1
     .else                         ; Micropolis
            ; DOS 2.7 - 901885-04 (8050 or 8250 with Micropolis drives)
;            .res 2
FILLER:     .byte $AA

LFFFA:      .byte $AA              ; Checksum 2
ROMCS:      .byte $95              ; Checksum 1
     .endif
    .elseif .def(DOS25)            ; DOS 2.5
     .ifdef DOS25B
            ; DOS 2.5 - 901483-03 (8050 with Micropolis drives)
;            .res 34
FILLER:     .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA

ROMCS:      .byte $1F              ; Checksum
     .else
            ; DOS 2.5 - 901483-02 (8050 - Micropolis ???)
;            .res 38
FILLER:     .byte $38, $A5, $45, $ED, $8F, $02, $A5, $44
            .byte $ED, $8E, $02, $B0, $1A, $A0, $05, $B9
            .byte $48, $03, $D1, $42, $D0, $13, $88, $10
            .byte $F6, $A0, $06, $B1, $42, $8D, $94, $02
            .byte $C8, $B1, $42, $8D, $95, $02

ROMCS:      .byte $9B              ; Checksum
     .endif

    .elseif .def(DOS1)            ; DOS 1.x
            ; DOS 1.x - 901466-01 or 901466-02 (2040 or 3040 with Shugart drives)
;            .res 9
FILLER:     .byte " IS ON AN"

            ; In this version (DOS 1.x) isn't a checksum!
    .else                         ; DOS 2.0 / DOS 2.1
            ; DOS 2.0 / DOS 2.1 - 901466-04 (4040 with Shugart drives)
;            .res 13
FILLER:     .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
            .byte $AA, $AA, $AA, $AA, $AA

ROMCS:      .byte $18              ; Checksum
    .endif

RES:        .word RESET
IRQ:        .word IRQH
