    .ifndef DOS27                  ; DOS 2.7 used DS disks
     .ifdef DOS25
MAXTRK      = 77                   ; Format 2C for 8050 / 8250 use 77 tracks
     .else
MAXTRK      = 35                   ; Format 2A for 2040 / 3040 / 4040 use 35 tracks
     .endif
    .endif

    .ifdef DOS1                    ; DOS 1.x uses direct access to FDC rom
STEPBT      = ANDD
    .endif

    .ifdef SPEEDOS                 ; Speedos don't use a buffer to execute the code, use zeropage and shared memory
L0870       = TRACK
SECTOR      = NEXTS                ; Number of current sector
BKPRTY      = WORK                 ; Header checksum
L0874       = STAB+4
L0875       = SPDVAR
L0876       = TAB3
L0877       = SECTR
L0878       = STAB+2
L0879       = L0026 ; (2 bytes as word)
L087B       = IP+1
L087C       = STAB+1
L087D       = CSECT
L087E       = STAB+3
L087F       = DSECT
    .endif

; ===> Format routine <===

FORMT:
    .ifdef DOS27
     .ifdef SPEEDOS
            lda #1                 ; Start with track 1
            sta L0870
            sta L087B
            sta L087C

            lda STAB
            sta L0878
     .else
            lda #$00
            sta L0877

            lda NSIDES
            sta L0878

            ldx DRIVE
            lda L0870
            sta L087B
            sta DRVTRK,X

            lda TAB3
            sta L0876
     .endif

            lda RIOT_PB
            ora #PB_HD_SEL
            sta RIOT_PB

L0521:      lda #3
            sta L087E

            lda #$30
            sta L087F
    .else
            lda FTNUM              ; Is format running?
            bpl LE042              ; Yes, do it

            sei
     .ifdef DOS25
            ldx DRIVE
            lda #ST_HEAD_STP
     .else
            lda #ST_SPEED_OK+ST_HEAD_STP+1
     .endif
            sta DRVST,X            ; Set drive status: head is stepping

     .ifdef DOS25
            lda #1
            sta DRVTRK,X           ; BUMP to track 1

            lda STEPBT,X
     .else
            lda #VB_TRK_S0B+VB_TRK_S0A+VB_TRK_S1B+VB_TRK_S1A
            and STEPBT,X
     .endif
            ora VIA_VB
            sta VIA_VB

     .ifdef DOS25
      .ifdef REVB
            lda #92                ; 92 Steps to BUMP for DOS 2.5 rev.
      .else
            lda #90                ; 90 Steps to BUMP for DOS 2.5
      .endif
     .else
            lda #140               ; 140 Steps to BUMP for 4040 (DOS 2.0 and DOS 2.1)
     .endif
            sta STEPS,X
            cli

; ===> Wait until head positioned and set pointers <===

LE01B:      lda STEPS,X
            bne LE01B

            tya
            asl A
            asl A
            asl A
            clc
            adc #<HDRS
            sta HDRPNT

     .ifdef DOS25
            ldy #10                ; Count down
     .else
            ldy #0                 ; Count up
     .endif
            sty ERRCNT             ; Reset error counter

     .ifdef DOS25
            ldy #1
     .else
            iny                    ; Y = 1
     .endif
            sty FTNUM              ; Track 1 to be formated

     .ifdef DOS25
            lda #<4800
            sta L08AA
            lda #>4800
            sta L08AA+1            ; Capacity of a track: 4800 < capacity < 2*4800
     .endif

     .ifdef DOS1
            iny                    ; Y = 2
            lda (HDRPNT),Y
            and #%11000000
            ora #1                 ; Track = 1
            sta (HDRPNT),Y
     .else
LE03A:      jsr L0847              ; Create header image for first sector
     .endif

            ldy JOBNUM
     .ifdef DOS1
            and #%00111111
LE03A:      sta TRKS,Y
            jmp START
     .else
            jmp (STARTI)
     .endif

; ===> Format a track <===

LE042:
     .ifdef DOS1
            cmp TRKS,Y
     .else
            ldy #2
            eor (HDRPNT),Y
     .endif
            bne LE03A

     .if !(.def(DOS1) || .def(DOS25))
            lda #0
            sta ERRCNT             ; Reset error counter
     .endif

; ===> Check write protect status <===

LE048:      sei
    .endif
    .ifdef DOS1
            lda #25
            sta DTRCK

L0548:      ldy #3
            lda #$00
            sta (HDRPNT),Y

            lda #IFR_CB1
            sta VIA_IER

            lda #PCR_WR_GCR
            sta VIA_PCR
    .else
     .ifdef DOS27
      .ifdef SPEEDOS
            jsr LFD6C
      .else
            jsr L0776
      .endif
     .else
            jsr L0847              ; Create header image for first sector
     .endif
    .endif

    .ifdef DOS27
            lda RIOT_PB
            and #PB_DRV_WPS
    .else
            lda #PB_DRV_WPS
            and RIOT_PB
    .endif
    .ifdef DOS1
            beq L0561

            lda #WRPROT
            jmp ERR
L0561:
    .else
     .ifdef SPEEDOS
            bne L0599              ; => Error #WRPROT
     .else
            jne L07F8              ; => Error #WRPROT
     .endif
    .endif

    .ifdef DOS1                    ; DOS 1.x 
; ===> Calculate and store header checksum <===

            ldy #3
            lda #$00
L0565:      eor (HDRPNT),Y
            dey
            bpl L0565
            ldy #4
            sta (HDRPNT),Y

            ldx #%11111111
            stx RIOT_EOUT
            sty WORK               ; WORK = 4

; ===> Write 4*256 bytes $00 <===

            inx                    ; X = 0
            ldy #0
L0577:      jsr OUT                ; Write character in .X to disk
            dey
            bne L0577
            dec WORK
            bne L0577

; ===> Write 256 bytes $00 <===

L0581:      ldx #$00
            ldy #0
L0585:      jsr OUT                ; Write character in .X ($00) to disk
            dey
            bne L0585
    .else
     .ifdef DOS27                  ; DOS 2.7
            jsr L07A6              ; Record 32*256 null bytes to current track

; ===> Write 5 bytes $FF as Synch <===

            lda #PCR_WR_RAW
            ldx #$FF
L053E:      bvc L053E
            clv
            stx RIOT_DOUT
            sta VIA_PCR
            ldy #5-1
            lda #PCR_RD_GCR
L0549:      bvc L0549
            clv
            dey
            bpl L0549
L054F:      sta VIA_PCR

            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB

      .ifdef SPEEDOS
            lda #%01000000+%00000001
      .else
            lda #%01000000
            ora #%00000001
      .endif
            sta VIA_ACR

            ldx #232               ; 232 us ???
            stx VIA_T1LL

            ldx #$03
            ldy #0
L0565:      lda VIA_VB
            bmi L0565
L0569:      lda VIA_VB
            bpl L0569
L056D:      stx VIA_TIMER

L056F:      lda VIA_T1LL
L0571:      lda VIA_VB
            bpl L0581
            lda VIA_IFR
            asl A
            bpl L0571
            iny
            bne L056F
            beq FMTERR

L057F:      lda VIA_T1LL
L0581:      lda VIA_VB
            bmi L058F
            lda VIA_IFR
            asl A
            bpl L0581
            iny
            bne L057F
            beq FMTERR

L058F:      cpy #$C5
            bcc FMTERR
            cpy #$CC
            bcc L059C

; ===> Format error <===

FMTERR:     lda #BADFMT
       .ifdef SPEEDOS
L0599:      jmp ERROR
       .else
            jmp L07F8
       .endif

; ===> Calculate Checksum <===

L059C:      lda L0870
            eor HDRS
            eor HDRS+1
            sta BKPRTY              ; Calculate checksum

            lda #0
            sta SECTOR              ; Start with sector number 0

            jsr L07BF
     .else
; ===> Check Speed <===

      .ifdef DOS25                 ; DOS 2.5
L0555:      jsr L088F              ; Record 50*256 sync bytes (raw $FF) to current track

            lda #PCR_WR_RAW
            jsr L087D
            lda #%00001111
            sta RIOT_DOUT
            lda #PCR_WR_GCR
            jsr L087D
            jsr L080A

            lda #98
            sta VIA_T1LL
            lda #0
            sta VIA_TIMER

            lda #%01000000
            sta VIA_ACR

            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB

            ldy #>0
            ldx #<0
L057F:      lda VIA_VB
            bmi L057F
L0583:      lda VIA_VB
            bpl L0583

L0587:      lda VIA_T1LL
L0589:      lda VIA_VB
            bpl L059B
            lda VIA_IFR
            asl A
            bpl L0589
            inx
            bne L0587
            iny
            bne L0587
            jmp FMTERR

L059B:      stx L08A8
            sty L08A8+1

            ldx #<0
            ldy #>0
L05A5:      lda VIA_T1LL
L05A7:      lda VIA_VB
            bmi L05B9
            lda VIA_IFR
            asl A
            bpl L05A7
            inx
            bne L05A5
            iny
            bne L05A5
            jmp FMTERR

L05B9:      sec
            txa
            sbc L08A8
            tax
            sta L08A6
            tya
            sbc L08A8+1
            tay
            sta L08A6+1

            bpl L05D7
            eor #$FF
            tay
            txa
            eor #$FF
            tax
            inx
            bne L05D7
            iny
L05D7:      tya
            bne L05DE
            cpx #$04
            bcc L05FA
L05DE:      asl L08A6
            rol L08A6+1
            clc
            lda L08A6
            adc L08AA
            sta L08AA
            lda L08A6+1
            adc L08AA+1
            sta L08AA+1
            jmp L0555

L05FA:      ldx #<0
            ldy #>0
L05FE:      lda VIA_VB
            bpl L060E
            bvc L05FE
            clv
            inx
            bne L05FE
            iny
            bne L05FE
            jmp FMTERR

L060E:      txa
            asl A
            sta COUNTL
            tya
            rol A
            sta COUNTH

            lda #%00000001
            sta VIA_ACR
      .else                        ; DOS 2.0
            jsr L07A6              ; Record 32*256 null bytes to current track

            ldx #$FF
            lda #PCR_WR_PLS
            jsr WRITEx             ; Send value in .X to output register
            sta VIA_PCR
            jsr WRITEx             ; Send value in .X to output register
            jsr WRITEx             ; Send value in .X to output register
            lda #PCR_WR_GCR
            jsr WRITEx             ; Send value in .X to output register
            ldx #OFFBYT
            jsr WRITEx             ; Send value in .X to output register
            sta VIA_PCR
            jsr L080A              ; Read four header bytes
            jsr L0827              ; Loop to time non-sync area
            jsr L0737              ; Read a data byte
            cmp #OFFBYT
            jne FMTERR
      .endif

; ===> Calculate track capacity <===

      .ifdef DOS25
            lda #18                ; Number of header bytes without OFF bytes (DOS 2.5)
      .else
            lda #17                ; Number of header bytes without OFF bytes (DOS 2.0)
      .endif
            clc
            adc GAP1               ; Size of gap after sector header
            sta DTRCK              ; = 17 or 18 + 9
            ldx SECTR              ; Highest sector number in current track
            ldy #>0                ; Calculate OFF bytes for gap
            lda #<0
L062A:      clc
            adc DTRCK              ; .A is now 26 or 27
            bcc L0630
            iny
L0630:      iny                    ; .Y is number of sector plus remainder from .A
            dex                    ; Leaving in (.Y,.A) the total of bytes
            bne L062A              ; Including the header, plus one extra byte in each

            eor #255
            sec
            adc #0                 ; This is then subtracted from the track capacity in (COUNTL)
            clc
            adc COUNTL             ; Resulting in a total of the gap bytes
            bcs L0642
            dec COUNTH
L0642:      tax
            tya
            eor #255
            sec
            adc #0
            clc
            adc COUNTH
            jmi FMTERR

; ===> Calculate gap size <===

            tay                    ; Total of gap bytes divided by the number of sectors
            txa
            ldx #0
L0656:      sec
            sbc SECTR              ; Highest sector number in current track
            bcs L065E
            dey
            bmi L0661
L065E:      inx
            bne L0656
L0661:      stx DTRCK
            cpx GAP2               ; Number of bytes between sectors
            jcc FMTERR

; ===> Calculate control value, write 32 or 50 * 256 null bytes <===

            clc
            adc SECTR              ; Highest sector number in current track
            sta CHKSUM             ; Control for header checksum
     .endif

            jsr L07A6              ; Record 32 or 50 * 256 null bytes to current track
    .endif

; ===> Record current sector header and data to disk <===

; ===> Write 4 bytes raw $FF as SYNC <===

L05B3:      lda #PCR_WR_RAW
    .ifdef DOS1
            stx WORK
            dex
    .else
            ldx #$FF
    .endif
    .if .def(DOS25)
L05B7:      bvc L05B7              ; First byte
            clv
            stx RIOT_DOUT
    .elseif .def(DOS1)
            jsr OUT                ; Write character in .X to disk
    .else
            jsr WRITEx             ; Send first value in .X to output register
    .endif
            sta VIA_PCR

    .ifdef DOS27
            ldx #HBID
    .else
     .if .def(DOS25)
L067Fx:     bvc L067Fx             ; Second byte
            clv
     .elseif .def(DOS1)
            jsr OUT                ; Write character in .X to disk
     .else
            jsr WRITEx             ; Send second value in .X to output register
     .endif
     .ifndef DOS1
      .ifdef DOS25
L0682:      bvc L0682              ; Third byte
            clv
      .else
            jsr WRITEx             ; Send third value in .X to output register
      .endif
     .endif
    .endif
            lda #PCR_WR_GCR
    .ifndef DOS1
     .ifdef DOS27
            ldy #3
     .endif
     .ifdef DOS25
L05C4:      bvc L05C4              ; Fourth byte or second to fourth byte
            clv
     .else
            jsr WRITEx             ; Send fourth value in .X to output register
     .endif
     .ifdef DOS27
            dey
            bne L05C4
     .endif
    .endif

; ===> Write header block ID ($08) <===

    .ifndef DOS27
            ldx #HBID
    .endif
    .if .def(DOS25)
L05CA:      bvc L05CA
            clv
            stx RIOT_DOUT
    .elseif .def(DOS1)
            jsr OUT                ; Write character in .X to disk
    .else
            jsr WRITEx             ; Send value in .X to output register
    .endif
            sta VIA_PCR

    .ifndef DOS1
; ===> Write BKPRTY (Checksum) <===

     .ifdef DOS27
            lda BKPRTY
     .endif
            ldx #$FF
     .ifndef DOS27
            lda BKPRTY
     .endif
L05D6:
     .ifdef DOS25
            bvc L05D6
            clv
     .else
            bit VIA_IFR
            bpl L05D6
     .endif
            sta RIOT_DOUT
     .ifndef DOS25
            bit VIA_DIN
     .endif
            eor SECTOR              ; Eleminate the sector number from checksum
            ldy SECTOR              ; Get the sector number
            inc SECTOR              ; Increment the sector number for the next sector

; ===> Write SECTOR (Sector number) <===

L05E4:
     .ifdef DOS25
            bvc L05E4
            clv
     .else
            bit VIA_IFR
            bpl L05E4
     .endif
            sty RIOT_DOUT
     .ifndef DOS25
            bit VIA_DIN
     .endif
            eor SECTOR              ; Include the next sector number and
            sta BKPRTY              ; prepare checksum for next sector
    .endif

    .ifdef DOS27
; ===> Write L0870 (Track number) <===

            lda L0870
L05F2:      bvc L05F2
            clv
            sta RIOT_DOUT

; ===> Write HDRS+1 and HDRS (ID2 and ID1) <===

            lda HDRS+1
            ldy HDRS
L05FD:      bvc L05FD
            clv
            sta RIOT_DOUT
L0602:      bvc L0602
            clv
            sty RIOT_DOUT
    .else

; ===> Write Track#, ID1 and ID2 from HDRS) <===

     .ifdef DOS1
            ldy #4
     .else
            ldy #2
     .endif
L06B3:      lda (HDRPNT),Y
L06B5:
     .ifdef DOS25
            bvc L06B5
            clv
     .else
            bit VIA_IFR
            bpl L06B5
     .endif
            sta RIOT_DOUT
    .ifndef DOS25
            bit VIA_DIN
    .endif
            dey
            bpl L06B3
    .endif

; ===> Write GAP1 count OFF Bytes <===

    .ifndef DOS1
     .ifdef DOS25
            lda #OFFBYT
     .else
            lda #$00
     .endif
     .ifdef DOS27
L0609:      bvc L0609
            clv
            sta RIOT_DOUT
            lda #PCR_WR_RAW
     .endif
            ldy GAP1
L0613:
     .ifdef DOS25
            bvc L0613
            clv
     .else
            bit VIA_IFR
            bpl L0613
     .endif
     .ifndef DOS27
            sta RIOT_DOUT
     .endif
     .ifndef DOS25
            bit VIA_DIN
     .endif
            dey
            bne L0613
     .ifndef DOS27
            lda #PCR_WR_RAW
     .endif

; ===> Write 5 bytes $FF as Synch <===

     .ifdef DOS25
L0619:      bvc L0619              ; First byte
            clv
            stx RIOT_DOUT
     .else
            jsr WRITEx             ; Send first value in .X to output register
     .endif
            sta VIA_PCR
     .ifdef DOS27
            ldy #3
     .endif
     .ifdef DOS25
L0622:      bvc L0622              ; Second byte
            clv
     .else
            jsr WRITEx             ; Send second value in .X to output register
     .endif
     .ifdef DOS27
            dey
            bne L0622              ; Third byte
     .endif
     .ifdef DOS25
L0628:      bvc L0628
            clv
     .else
            jsr WRITEx             ; Send third value in .X to output register
     .endif
            lda #PCR_WR_GCR
     .ifndef DOS27
      .ifdef DOS25
L06DB:      bvc L06DB              ; Fourth byte
            clv
      .else
            jsr WRITEx             ; Send fourth value in .X to output register
      .endif
     .endif
            ldx #BID
     .ifndef DOS27
      .ifdef DOS25
L06E0:      bvc L06E0              ; Fifth byte (only DOS 2.5 and later)
            clv
      .endif
     .endif

; ===> Write $07 as Block ID <===

     .ifdef DOS25
L062F:      bvc L062F
            clv
            stx RIOT_DOUT
     .else
            jsr WRITEx             ; Send value in .X to output register
     .endif
            sta VIA_PCR

; ===> Write 256 bytes $00 as empty block data <===

     .ifdef SPEEDOS
            ldy #$00
     .elseif .def(DOS27)
            ldy #$FF
            lda L0873
     .else
            ldy #$00
      .ifdef DOS25
            lda #$00
      .else
            ldx #$00
      .endif
L0640:
     .endif
L063B:
     .ifdef DOS25
            bvc L063B
            clv
      .ifdef SPEEDOS
            sty RIOT_DOUT
      .else
            sta RIOT_DOUT
      .endif
     .else
            bit VIA_IFR
            bpl L063B
            stx RIOT_DOUT
            bit VIA_DIN
     .endif
     .ifdef DOS27
L0640:      bvc L0640
            clv
     .endif
            dey
            bne L0640
     .ifdef DOS27
      .ifndef SPEEDOS
L0646:      lda #$00
L0648:      bvc L0648
            clv
            sta RIOT_DOUT
      .endif
     .endif
    .endif

; ===> Switch from write to read and delay L0874 bytes <===
; ===> Write DTRCK bytes $00 as gap between sectors <===

    .ifdef DOS27
            ldy L0874
            ldx #PCR_RD_GCR
    .else
            ldy DTRCK
     .ifdef DOS25
            iny
     .else
      .ifndef DOS1
            jsr WRITEx             ; Send value in .X to output register
      .endif
     .endif
    .endif
L0652:
    .if .def(DOS25)
            bvc L0652
            clv
    .elseif .def(DOS1)
            jsr OUT                ; Write character in .X to disk
    .else
            bit VIA_IFR
            bpl L0652
            stx RIOT_DOUT
            bit VIA_DIN
    .endif
            dey
    .if .def(DOS25) || .def(DOS1)
            bne L0652
    .else
            bpl L0652
    .endif
    .ifdef DOS27
L0658:      bvc L0658
            clv
    .endif

; ===> Now writing a sector is finish, next one? <===

    .ifdef DOS1
            ldy #3
            lda (HDRPNT),Y
            sta WORK
            clc
            adc #1                 ; Increment sector#
            sta (HDRPNT),Y
    .else
            lda SECTOR
    .endif
    .ifdef DOS27
            cmp L0877              ; All sectors written?
    .else
            cmp SECTR
    .endif
    .ifdef DOS1
            beq L05D1

            iny
            eor WORK
            eor (HDRPNT),Y
            sta (HDRPNT),Y         ; Calculate new header checksum for new sector#
            jmp L0581

L05D1:
    .else
            jne L05B3              ; Write next sector
    .endif

    .ifdef DOS25
     .ifdef DOS27
; ===> ??? <===

            stx VIA_PCR

            lda #100
            sta VIA_T1LL
            lda #0
            sta VIA_TIMER

L0670:      lda VIA_IFR
            asl A
            bpl L0670

            lda #$10
            sta L087D

            lda #1
            sta SECTOR

L067F:      ldx #200               ; Maximum of 200 tries to read the sync with correct header block ID
     .else
            jsr L080A

            lda #80
            sta WORK               ; Maximum of 80 tries to read the sync with correct header block ID

            dec SECTOR             ; Now decrement the sector#
     .endif
L0681:
     .ifdef DOS27
            jsr L0827
     .else
L0713:      jsr L0827              ; Loop to time non-sync area
L0716:      bvc L0716
            clv
            lda VIA_DIN            ; Read ID after SYNC
     .endif
            cmp #HBID              ; Is this a header block ID?
            beq L0690              ; Yes => all okay
L0688:
     .ifdef DOS27
            dex
     .else
            dec WORK               ; Decrement try count
     .endif
            bne L0681              ; Is there an other try possible
L068B:
     .ifdef DOS27
            lda #NOHDR
      .ifdef SPEEDOS
            bne LF98B              ; No => #NOHDR error
      .else
            jmp L07F8              ; No => error
      .endif
     .else
            jmp FMTERR             ; No => error
     .endif

L0690:      bvc L0690
            clv
            lda VIA_DIN            ; Read header checksum
     .ifdef DOS27
            sta BKPRTY
     .endif
L0698:      bvc L0698
            clv
            lda VIA_DIN            ; Read sector#
            cmp SECTOR
            bne L0681
    .else
     .ifdef DOS1
            jsr OFF
     .else
            jsr L080A
     .endif
    .endif

    .ifdef DOS1
            ldx #250
            jsr WATCH
            bvs L05E1

            dec DTRCK
            beq L05FB
            bne L05F0

L05E1:      ldx #40
            jsr WATCH
            bvc L05F3

            inc DTRCK
            lda DTRCK
            cmp #$30
            beq L05FB
L05F0:      jmp L0548

L05F3:      lda ERRCNT
            cmp #$30
            bne L0604

            asl JOB

L05FB:      lda #$FF
            sta FTNUM
            lda #BADFMT
            jmp ERROR

L0604:      lda #$80
            sta JOB
            lda #$00
            ldy #3
            sta (HDRPNT),Y
L060E:      jsr SRCH
            ldy #3
            lda (HDRPNT),Y
            clc
            adc #$01
            sta (HDRPNT),Y
            cmp SECTR
            bne L060E

            asl JOB
    .else
     .ifdef DOS27
L06A2:      eor BKPRTY
            ldy #3
L06A7:      bvc L06A7
            clv
            eor VIA_DIN
            dey
            bne L06A7

L06AF:      tay
      .ifdef SPEEDOS
            bne L06B2
      .else
            beq L06B7

L06B2:      lda #BADHCH
L06B4:      jmp L07F8
      .endif

L06B7:      jsr L0827
     .else
            lda #0
            sta SECTOR

      .ifdef DOS25
            jsr L0827              ; Loop to time non-sync area
L073D:      bvc L073D
            clv
            lda VIA_DIN            ; Read ID after SYNC
      .endif
     .endif
     .ifdef DOS25
      .ifdef SPEEDOS
            eor #BID
      .else
            cmp #BID               ; Is this a data block ID?
      .endif
      .ifdef DOS27
            beq L06C3
L06BE:      lda #NODBLK
       .ifdef SPEEDOS
            .byte $2C

L06B2:      lda #BADHCH
            .byte $2C

LF989:      lda #BADBCH
       .endif
LF98B:      jmp L07F8

L06C3:
      .else
            jne FMTERR             ; No => error
      .endif
      .ifdef DOS27
       .ifdef SPEEDOS
            tay
       .else
            ldy #$00
            tya
       .endif
      .else
            ldx #$00
      .endif
L06C6:      bvc L06C6
            clv
      .ifdef DOS27
            eor VIA_DIN
            dey
      .else
            dex
      .endif
            bne L06C6
L06CE:      bvc L06CE
            clv
      .ifdef DOS27
            eor VIA_DIN
       .ifdef SPEEDOS
            bne LF989
       .else
            beq L06DA
            lda #BADBCH
            jmp L07F8
L06DA:
       .endif

            inc SECTOR
            lda SECTOR
            cmp L0877
            beq *+5
            jmp L067F

            cmp #$01
            beq L0718
      .else
            ldy DTRCK
L0756:      bvc L0756
            clv
            dey
            bne L0756
      .endif
     .endif

; ===> Check header ID <===

L06EC:      jsr L0827              ; Loop to time non-sync area

     .ifndef DOS27
      .ifdef DOS25
L075F:      bvc L075F
            clv
            lda VIA_DIN            ; Read ID after SYNC
      .else
            jsr L0737              ; Read a data byte
      .endif
     .endif
            cmp #HBID              ; Is this a header block ID?
     .ifdef DOS27
            bne L0706

L06F3:      ldy L0879+1
            bne L0700
L06F8:      ldy L0879
            cpy L0875
            bcc L070C
L0700:      inc L0874
      .ifdef SPEEDOS
            .byte $2C
      .else
            jmp L059C
      .endif

L0706:      dec L0874
L0709:      jmp L059C

L070C:      ldy #$01
            sty L0877
            dey                    ; Y := 0
            sty SECTOR
            jmp L067F

L0718:      dec L0878
            beq L0732

            lda RIOT_PB
            and #255-PB_HD_SEL
            sta RIOT_PB

            ldx L0876
            dex
            txa
            clc
            adc L0870
            sta L0870

      .ifdef SPEEDOS
            bne L0709
      .else
            jmp L059C
      .endif

L0732:      lda RIOT_PB
            ora #PB_HD_SEL
            sta RIOT_PB
      .ifdef SPEEDOS
            lda STAB
      .else
            lda NSIDES
      .endif
            sta L0878
      .ifdef SPEEDOS
            ldx DRIVE
            ldy L087B
            iny
            cpy L0876
      .else
            ldx L087B
            inx
            cpx L0876
      .endif
            beq FMTEND
L0747:
      .ifdef SPEEDOS
            sty DRVTRK,X
            sty L0870
            sty L087B
      .else
            stx L0870
            stx L087B
      .endif
            ldy #$FF
      .ifndef SPEEDOS
            lda LFFFA
            bne L0755
            dey
L0755:
      .endif
      .ifdef SPEEDOS
            jsr LFD7A
      .else
            jsr L0782
      .endif
            lda #3
            sta L087E
      .ifdef SPEEDOS
            bne L0709
      .else
            jmp L059C
      .endif

      .ifndef SPEEDOS
            ldx L0876
            dex
            stx L0876
            jmp L0521
      .endif

FMTEND:
     .else
L0766:      beq L0778
      .ifdef DOS25
            jmp FMTERR             ; No => error
      .endif

; ===> Format error <===

FMTERR:     lda #BADFMT

; ===> Permit ten tries before aborting <===

L07F8:      cli
      .ifdef DOS25
            dec ERRCNT             ; Decrement error count
            jeq L0803              ; Too many errors detected => Abort with error# BADFMT
      .else
            inc ERRCNT
            ldy #10
            cpy ERRCNT
            beq L0803
      .endif

            jmp LE048              ; Try same track again

; ===> Read and store header checksum <===

L0778:
      .ifdef DOS25
            bvc L0778
            clv
            lda VIA_DIN            ; Read header checksum
      .else
            jsr L0737
      .endif
            sta BKPRTY             ; And store it

; ===> Read SECTOR (Sector#)

      .ifdef DOS25
L0780:      bvc L0780
            clv
            lda VIA_DIN            ; Read sector#
      .else
            jsr L0737
      .endif
            cmp SECTOR             ; Is the sector# correct?
      .ifdef DOS25
            beq *+5
            jmp FMTERR             ; no => error
      .else
            bne FMTERR             ; no => error
      .endif
            eor BKPRTY             ; calculate chacksum

; ===> Read Track#, ID1 and ID2 <===

            ldy #2
L0792:
      .ifdef DOS25
            bvc L0792
            clv
      .else
            bit VIA_IFR
            bpl L0792
      .endif
            eor VIA_DIN            ; Read three bytes and calculate checksum
            dey
            bpl L0792

; ===> Check checksum <===

            tay                    ; Is checksum correct?
      .ifdef DOS25
            beq *+5
            jmp FMTERR             ; No => error
      .else
            bne FMTERR             ; No => error
      .endif
            inc SECTOR             ; Increment the readed sector header

; ===> Read $07 as block ID <===

            jsr L0827              ; Loop to time non-sync area
      .ifdef DOS25
L07A6a:     bvc L07A6a
            clv
            lda VIA_DIN            ; Read ID after SYNC
      .else
            jsr L0737
      .endif
            cmp #BID               ; Is this a data block ID?
      .ifdef DOS25
            beq *+5
            jmp FMTERR             ; No => error
      .else
            bne FMTERR             ; No => error
      .endif
            ldy #$00

; ===> Calculate checksum for data sector <===

L07B4:
      .ifdef DOS25
            bvc L07B4
            clv
      .else
            bit VIA_IFR
            bpl L07B4
      .endif
            lda VIA_DIN            ; Read data bytes
      .ifdef DOS25
            beq L07BE              ; Is the readed data byte 0?
            jmp FMTERR             ; No => error
L07BE:
      .else
            bne FMTERR
      .endif
            dey
            bne L07B4
L07C1:
      .ifdef DOS25
            bvc L07C1
            clv
            lda VIA_DIN            ; Read checksum, is the readed data byte 0?
            beq *+5
            jmp FMTERR             ; No => error
      .else
            jsr L0737              ; Read a data byte (checksum), is the readed data byte 0?
            bne FMTERR             ; No => error
      .endif

; ===> Now reading a sector is finish, next one? <===

            lda SECTOR
            cmp SECTR              ; All sectors in this track readed?
            bne L06EC              ; No, loop to the next one

; ===> Now reading a whole track finish, measure the gap from last sector to first sector <===

            jsr L0827              ; Loop to time non-sync area
            lda COUNTH
            beq *+5
            jmp FMTERR

            lda CHKSUM             ; Control for header checksum
            sec
            adc DTRCK              ; Size of gap
            sec
            sbc COUNTL             ; Minus the total bytes read
            bpl L07EE
            eor #255
            sec
            adc #0
L07EE:
      .ifdef DOS25
            cmp #50                ; Between -50 and +50 for 8050/8250?
      .else
            cmp #28                ; Between -28 and +28 for 4040?
      .endif
      .ifdef DOS25
            jcs FMTERR             ; No => error
      .else
            bcc *+5
            jmp FMTERR             ; No => error
      .endif
     .endif
    .endif

    .ifndef DOS27
; ===> Find out if this is last track to be formatted <===

            inc FTNUM              ; Track currently being formatted
     .ifdef DOS1
            lda FTNUM
            cmp #MAXTRK+1          ; All tracks formatted?
            beq L0803

            ldy #2
            lda (HDRPNT),Y
            and #%11000000
            ora FTNUM
            sta (HDRPNT),Y

            ldx DRIVE
            inc TRKS,X
            jmp START

     .else
            cli
            lda #MAXTRK+1
            cmp FTNUM              ; All tracks formatted?
            jne LE03A              ; No => do next track
     .endif
    .endif

; ===> Exit with OK flag <===

    .ifdef SPEEDOS
            lda #%10000000+GOODJ
            jmp ERROR
    .else
     .ifdef DOS27
            ldx DRIVE
            lda L087B
            sta DRVTRK,X
     .endif
     .ifndef DOS1
            lda #GOODJ
     .endif

; ===> Exit with format done flag, error number in .A <===

     .ifndef DOS27
L0803:
      .ifdef DOS1
            lda #$FF
            sta FTNUM              ; Track currently being formatted
      .else
            ldy #$FF
            sty FTNUM              ; Track currently being formatted
      .endif
     .endif
     .ifdef DOS1
            jmp DONE
     .else
            jmp (DONEI)
     .endif
    .endif

    .ifndef DOS1
     .ifndef SPEEDOS
      .ifdef DOS27
; ===> ??? <===

L0776:      sei
            ldx DRIVE
            lda STEPBT,X
            ora VIA_VB
            sta VIA_VB
            ldy #92
L0782:      ldx DRIVE
            lda #%01000000
            sta DRVST,X
            tya
            sta STEPS,X
            lda L04EE
            sta STPCNT,X
L0791:      sei
            lda MTRCLK
            sta MTRTM,X
            dec MTRTM,X
            cli
            lda STEPS,X
            bne L0791
L079E:      lda STPCNT,X
            bne L0791
L07A2:      sei
            rts
      .else
; ===> <===

L080A:
       .ifdef DOS25
            bvc L080A
            clv
L080D:      bvc L080D
            clv
       .else
            jsr WRITEx             ; Send value in .X to output register
L0710:      bit VIA_IFR
            bpl L0710
       .endif
            lda #PCR_RD_GCR
            sta VIA_PCR
       .ifdef DOS25
            ldx #10
L0816:      bvc L0816
            clv
       .else
            lda #IFR_IRQ+IFR_CB1+IFR_CA1
            sta VIA_IER
            ldx #4-1
L0816:      jsr L0737
            bit VIA_VB
       .endif
            dex
            bne L0816
            rts

       .ifndef DOS25
; ===> Switch to write mode <===

L0727:      ldy #IFR_CB1
            sty VIA_IER

            sta VIA_PCR
            rts

; ===> Send value in .X to output register <===

WRITEx:     bit VIA_IFR
            bpl WRITEx
            stx RIOT_DOUT
            bit VIA_DIN
            rts

; ===> Read a data byte <===

L0737:      bit VIA_IFR
            bpl L0737
            lda VIA_DIN
            rts
       .endif

; ===> Loop to time non-sync area <===

L0827:
       .ifdef DOS25
            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB
       .endif
            ldy #$00
            sty COUNTH
L0828:
       .ifdef DOS25
            lda VIA_VB
            bpl L083A
            bvc L0828
            clv
       .else
            bit RIOT_PB
            bvc L083A
            bit VIA_IFR
            bpl L0828
            bit VIA_DIN
            bit VIA_VB
       .endif
            iny
            bne L0828
            inc COUNTH
       .ifdef DOS25
            bne L0828
            jmp FMTERR
       .else
            jeq FMTERR
            jmp L0828
       .endif

L083A:      sty COUNTL
       .ifdef DOS25
            lda VIA_VB
            ora #VB_PLLSYN
            sta VIA_VB
            lda VIA_DIN
            clv
       .else
            bit VIA_VB
            bit VIA_DIN
       .endif
            rts

; ===> Create header image for first sector <===

L0847:      ldy #2
            lda FTNUM
            sta (HDRPNT),Y         ; Store as track#
            lda #0                 ; First sector is sector 0
            sta SECTOR
            iny                    ; Y = 3
            sta (HDRPNT),Y         ; Store as sector#

; ===> Calculate checksum <===

L0855:      eor (HDRPNT),Y         ; Calculate checksum
            dey
            bpl L0855
            sta BKPRTY
            ldy #4
            sta (HDRPNT),Y         ; Store checksum
            rts

; ===> Record 32/50*256 null bytes to current track <===

L07A6:      ldx #$00
            ldy #0
       .ifdef DOS25
            lda #50                ; 50 * 256 bytes for 8050/8250
       .else
            lda #32                ; 32 * 256 bytes for 4040
       .endif
            sta COUNTH
            lda #PCR_WR_GCR
       .ifdef DOS25
            sta VIA_PCR
       .else
            jsr L0727              ; Switch to write mode
       .endif
L086F:
       .ifdef DOS25
            bvc L086F
            clv
            stx RIOT_DOUT
       .else
            jsr WRITEx             ; Send value in .X to output register
       .endif
            dey
            bne L086F
            dec COUNTH
            bne L086F
            rts

       .ifdef DOS25
; ===> ??? <===

L087D:      ldx L08AA
            ldy L08AA+1
            sta VIA_PCR
L0885:      bvc L0885
            clv
            dex
            bne L0885
            dey
            bpl L0885
            rts

; ===> Record 50*256 sync bytes (raw $FF) to current track <===

L088F:      ldx #$FF
            ldy #0
            lda #50
            sta COUNTH
            lda #PCR_WR_RAW
            sta VIA_PCR
            jmp L086F
       .endif
      .endif

      .ifdef DOS1
STEPBT      = ANDD
      .else
; ===> Step bits drive 0/1 <===

STEPBT:     .byte %00001100,%00000011
      .endif
     .endif

     .ifdef DOS27
; ===> Record 50*256 null bytes to current track <===

L07A6:      ldx #$00               ; Null byte
            ldy #0                 ; Counter low
            lda #PCR_WR_GCR
L07AC:      bvc L07AC
            clv
            stx RIOT_DOUT
            sta VIA_PCR
            ldx #50                ; Counter high
L07B5:      bvc L07B5
            clv
            dey
            bne L07B5
            dex
            bne L07B5
            rts

; ===> <===

L07BF:      ldx #4
            lda L0870
            cmp TAB3
            bcs L07D1
L07C9:      cmp TAB3-1,X
            dex
            bcs L07C9
            bcc L07D7

L07D1:      cmp TAB3+3,X
            dex
            bcs L07D1

L07D7:      lda TAB1,X
            sta L0877
            cmp L087C
            beq L07F7
            sta L087C

            lda GAP2
            sta L0874

            lda RIOT_PB
            and #255-PB_DRV_DS
            sta RIOT_PB

            txa
            asl A
            ora RIOT_PB
            sta RIOT_PB

L07F7:      rts

; ===> ??? <===

L07F8:
      .ifndef SPEEDOS
            cmp #BADFMT
            beq L0824
            cmp #WRPROT
            beq L0824
      .endif
            dec L087F
            beq L0818
            dec L087D
            jne L0681

            dec L087E
            beq L0818
      .ifndef SPEEDOS
            lda L0870
      .endif
            jmp L059C

L0818:      ldx L0870
            ldy SECTOR
            stx HDRS+2
            sty HDRS+3

L0824:
      .ifdef SPEEDOS
            jmp ERROR
      .else
            jmp (DONEI)
      .endif

; ===> <===

L0827:
      .ifdef DOS25
            lda VIA_VB
            and #255-VB_PLLSYN
            sta VIA_VB
      .endif
            ldy #$00
      .ifdef DOS27
            sty L0879
      .endif
            sty L0879+1

L0835:      lda VIA_VB
            bpl L084D
            bvc L0835
            clv
      .ifdef DOS27
            inc L0879
      .else
            iny
      .endif
            bne L0835
            inc L0879+1
            bne L0835

      .ifdef DOS27
            pla
            pla
            lda #NOSYNC
      .endif
            jmp L07F8

L084D:      ldy SYNDLY
L0850:      lda VIA_VB
      .ifdef DOS27
            bmi L0835
            dey
            bne L0850
      .endif

      .ifdef SPEEDOS
            jmp LFFCB
      .else
L0857:      ora #VB_PLLSYN
            sta VIA_VB
            lda VIA_DIN
            clv
       .ifdef DOS27
L085E:      bvc L085E
            clv
            lda VIA_DIN
       .endif
            rts
      .endif

      .ifdef DOS27
       .ifndef SPEEDOS
;            .res 12
             .byte $AA, $AA, $AA, $AA, $AA, $AA, $AA, $AA
             .byte $AA, $AA, $AA, $AA
       .endif
      .endif
     .endif

; ===> Variables for formatting rountine <===

     .ifdef DOS27
      .ifndef SPEEDOS
L0870       := *+ 0
SECTOR      := *+ 1                ; Number of current sector
BKPRTY      := *+ 2                ; Header checksum
L0873       := *+ 3 ;Fill Byte for empty sector data
L0874       := *+ 4 ;DTRCK (GAP2)
L0875       := *+ 5 ;Compare with L0879
L0876       := *+ 6 ;(TAB3) Compare with L089B
L0877       := *+ 7 ;SECTR (TAB1,X)
L0878       := *+ 8 ;Head Counter
L0879       := *+ 9 ; (2 bytes as word)
L087B       := *+11 ;FTNUM?
L087C       := *+12 ;FTNUM?
L087D       := *+13
L087E       := *+14
L087F       := *+15
      .endif
     .else
SECTOR      := *+ 0                ; Number of current sector
BKPRTY      := *+ 1                ; Header checksum
CHKSUM      := *+ 2                ; Control for header checksum
COUNTH      := *+ 3                ; Timer/Counter high
COUNTL      := *+ 4                ; Timer/Counter low
      .ifdef DOS25
L08A6       := *+ 5; (2 bytes as word)
L08A8       := *+ 7; (2 bytes as word)
L08AA       := *+ 9; (2 bytes as word)
      .endif
     .endif
    .endif
