; ===> Interrupt routine <===

IRQH:       pha                    ; 48
            txa                    ; 8A
            pha                    ; 48
    .ifndef DOS27
            lda TICK               ; AD 00 04
    .else
     .if .def(SPEEDOS)             ; Speedos
            lda TICK
     .elseif .def(PANA) || .def(TANDON) ; Matsushita (Panasonic) or Tandon
            lda #4-1               ; in 4096 us next interrupt for Matsushita or Tandon drives (244 Hz)
     .elseif .def(MPI)             ; MPI
            lda #6-1               ; in 6144 us next interrupt for MPI drives (183 Hz)
     .else                         ; Micropolis
            lda #8-1               ; in 8192 us next interrupt for Micropolis drives (122 Hz)
     .endif
            sta RIOT_T8
    .endif
            sta RIOT_T8            ; 85 8F

            inc CLOCK              ; E6 00
    .ifdef DOS25
            lda CLOCK
            and #$0F
            bne LFF24
            inc MTRCLK
LFF24:      ldx #$01
            lda RIOT_PB
            pha
            ora #PB_DRV_SEL
LFF2B:      sta RIOT_PB
            lda RIOT_PB
            and #PB_DRV_WPS
            cmp LWPT,X
            sta LWPT,X
            beq L898
            lda #$01
     .if .def(SPEEDOS) && .def(PANA) && .def(REVB)
            jsr LFBDA
     .else
            sta WPSW,X
     .endif
    .else
            ldx #$01               ; A2 01
    .endif

; ===> Select drive 1 <===

L898:       lda MTRCLK
            cmp MTRTM,X            ; D5 01
            bne L921               ; D0 16
    .ifndef DOS25
            lda ANDC,X             ; BD ED FF
    .endif
            asl DRVST,X            ; 16 03
    .ifdef DOS25                   ; For DOS 2.5 and later (QD drives)
            bcs LFFA8
            lda ANDC,X
            ora VIA_VB
            sta VIA_VB
LFFA8:
     .ifdef SFD
            lda MTRCLK
            sta MTRTM,X
            dec MTRTM,X
     .else
            jsr LFFC6
     .endif
    .else                          ; For DOS 2.1 and earlier (DD drives)
            bit VIA_VB             ; 24 40
            bcs LFFA8              ; B0 08
            sec                    ; 38
     .ifdef DOS1
            lda ANDE,X             ; BD F1 FF
     .else
            lda ANDC,X
     .endif
            eor VIA_VB             ; 45 40
            sta VIA_VB             ; 85 40
LFFA8:      bne L941               ; D0 01
    .endif
    .ifdef SPEEDOS
            lsr DRVST,X
    .else
            clc
L941:       ror DRVST,X
    .endif

; ===> Activate head stepping <===

L921:
    .ifdef SPEEDOS
            lda MTRCLK
            cmp L000A,X
            bne LFB81
            lda DRVST,X
            and #255-ST_MOTOR_ON
            sta DRVST,X
LFB81:
    .endif
            lda STEPS,X
            bne L911
    .ifdef DOS25
     .ifdef DOS27                  ; For DOS 2.7 only
            lda STPCNT,X
            beq LFF90
            dec STPCNT,X
            bpl LFFD6
LFF90:
      .ifdef SPEEDOS
       .if .def(MPI) || .def(PANA) ; For MPI and Matsushita (Panasonic) drives
            sta PHASE,X
       .else                       ; For Micropolis drives
            lda PHASE,X
            bne LFB9C
       .endif
      .else
            sta PHASE,X
      .endif
     .else                         ; For DOS 2.5 only
            lda PHASE,X
            beq LFF6D
            sta STEPS,X
            pha
            lda #$01
            sta STPCNT,X
            lda #$00
            sta PHASE,X
            pla
            bne L911
     .endif
    .endif

LFF6D:      lda DRVST,X            ; B5 03
            and #255-ST_HEAD_STP
            sta DRVST,X            ; 95 03
    .ifdef DOS27
            jmp LFFD6
    .else
            jmp LFFE4              ; 4C E4 FF
    .endif

; ===> Step head to next track <===

    .ifdef SPEEDOS
     .if !(.def(MPI) || .def(PANA)); For Micropolis drives only
LFB9C:      sta STEPS,X
            pha
            lda #1                 ; Try only one step to read near a track
            sta STEPS2,X
            lsr A                  ; A = 0
            sta PHASE,X
            pla
     .endif
    .endif

L911:       asl A                  ; 0A
    .if .def(MPI) || .def(PANA) || .def(TANDON)    ; For MPI and Matsushita (Panasonic) drives 
            lda VIA_VB
            and ANDA,X
            sta COW

            lda VIA_VB
            and ANDD,X
            bcs L912

            dec STEPS,X
            eor ANDD,X
            jmp LFFB7

L912:       inc STEPS,X

LFFB7:      eor ANDB,X
            beq LFFC4
            cmp ANDD,X
            beq LFFC4
            eor ANDD,X
LFFC4:
    .else                          ; For Shugart or Micropolis drives
     .ifdef DOS25                  ; For Micropolis drives
      .ifdef DOS27
            lda #4                 ; micropolis use 4 steps for 1 track
      .endif
            bcs L912
      .ifdef DOS27
            dec STEPS2,X
      .else
            dec STPCNT,X
      .endif
            bne LFF84
      .ifdef DOS27
            sta STEPS2,X
      .else
            lda STPTRK
            sta STPCNT,X
      .endif
            dec STEPS,X
LFF84:      lda VIA_VB
            and ANDD,X
            sec
            sbc ANDB,X
            jmp LFFD5

L912:
      .ifdef DOS27
            dec STEPS2,X
      .else
            dec STPCNT,X
      .endif
            bne LFF9B
      .ifdef DOS27
            sta STEPS2,X
      .else
            lda STPTRK
            sta STPCNT,X
      .endif
            inc STEPS,X
LFF9B:      lda VIA_VB
            and ANDD,X
            clc
            adc ANDB,X
     .else                         ; For Shugart drives (SA-400)
            lda VIA_VB             ; A5 40
            and ANDA,X             ; 3D D4 FC
            sta COW                ; 85 07

            lda VIA_VB             ; A5 40
            and ANDD,X             ; 3D EF FF
            bcs L912               ; B0 07

            dec STEPS,X            ; D6 05
            adc ANDB,X             ; 7D EB FF
            bne LFFD5              ; D0 05

L912:       inc STEPS,X            ; F6 05
            sbc ANDB,X             ; FD EB FF
     .endif
LFFD5:      and ANDD,X             ; 3D EF FF
     .ifdef DOS25                  ; for DOS 2.5 and later
            sta COW
            lda VIA_VB
            and ANDA,X
     .endif
    .endif
            ora COW                ; 05 07
            sta VIA_VB             ; 85 40
    .ifdef DOS25                   ; for DOS 2.5 and later
LFFD6:      lda RIOT_PB
            and #255-PB_DRV_SEL
LFFE4:      dex
     .if (*+2) - LFF2B > 128
            jpl LFF2B
     .else
            bpl LFF2B
     .endif

            lsr CFLG2
            pla
            sta RIOT_PB
    .else                          ; for DOS 2.1 and earlier
            lda DRVST,X            ; B5 03
            bmi LFFE4              ; 30 04
            lda MTRCLK
            sta MTRTM,X            ; 95 01
LFFE4:      dex                    ; CA
            bpl L898               ; 10 AA
    .endif
            pla                    ; 68
            tax                    ; AA
            pla                    ; 68
            rti                    ; 40
