@ECHO OFF

REM Assemble all FDC roms
REM ---------------------
REM (C) 2014 by Martin Hoffmann-Vetter
REM Assemble all known FDC roms from the universable source code.

IF NOT EXIST tmp MD tmp
IF NOT EXIST out MD out

FOR %%i IN (901466-01 901466-04 901483-02 901483-03 901884-01 901885-04 901869-01 251474-01b 251257-02a fdc-2.7b 2516_cbm8250_edotronik speeddos-fdc-f800) do (
	SET file=%%i
	ECHO Processing %%i

	REM Assemble the source to a temporary object code and generate a listing file
	SET pass=Assemble
	bin\ca65 -D fdc %%i.s -o tmp\%%i.o -l
	IF ERRORLEVEL 1 GOTO Error

	REM Link the binaries from the object code and generate a label and memory map file
	Set pass=Linker
	IF "%%i" == "fdc-2.7b"               bin\ld65 -C fdc2k.cfg tmp\%%i.o -o out\%%i.bin -Ln tmp\%%i.lbl -m tmp\%%i.map
	IF "%%i" == "2516_cbm8250_edotronik" bin\ld65 -C fdc2k.cfg tmp\%%i.o -o out\%%i.bin -Ln tmp\%%i.lbl -m tmp\%%i.map
	IF "%%i" == "speeddos-fdc-f800"      bin\ld65 -C fdc2k.cfg tmp\%%i.o -o out\%%i.bin -Ln tmp\%%i.lbl -m tmp\%%i.map
	IF NOT "%%i" == "fdc-2.7b" IF NOT "%%i" == "2516_cbm8250_edotronik" IF NOT "%%i" == "speeddos-fdc-f800" bin\ld65 -C fdc.cfg tmp\%%i.o -o out\%%i.bin -Ln tmp\%%i.lbl -m tmp\%%i.map
	IF ERRORLEVEL 1 GOTO Error

	REM Copy the format and init routines from BC roms
	SET pass=Compare BC
	IF "%%i" == "901466-01" MOVE /Y BC.bin BC_DOS_1.x.bin
	IF "%%i" == "901466-01" FC /b BC_DOS_1.x.bin org\901468-06_part.bin
	IF "%%i" == "901466-04" MOVE /Y BC.bin BC_DOS_2.0.bin
	IF "%%i" == "901466-04" FC /b BC_DOS_2.0.bin org\901468-11_part.bin
	IF "%%i" == "901483-03" MOVE /Y BC.bin BC_DOS_2.5.bin
	IF "%%i" == "901483-03" FC /b BC_DOS_2.5.bin org\901482-03_part.bin
	IF "%%i" == "251257-02a" MOVE /Y BC.bin BC_DOS_2.7.bin
	IF "%%i" == "251257-02a" FC /b BC_DOS_2.7.bin org\901887-01_part.bin
	IF ERRORLEVEL 1 GOTO Error

	REM Check the DC checksum (Not valid for DOS 1.x DC rom 901466-01)
	SET pass=Checksum
	IF NOT "%%i" == "901466-01" cscript -nologo bin\csum.vbs out\%%i.bin
	IF ERRORLEVEL 1 GOTO Error

	REM Compare the binaries
	SET pass=Compare DC
	fc /b out\%%i.bin org\%%i.bin
	IF ERRORLEVEL 1 GOTO Error
)

SET file=
SET pass=
ECHO All done without errors!
GOTO End

:Error
ECHO %pass% error in %file% found!
PAUSE

:End